//------------------------------------------------
// Copyright (C) 2012 Bo Xu <xuxbob@gmail.com>
//------------------------------------------------

#include "bppengine.h"
#include "def.h"
#include "cstylebppctlloader.h"
#include <QDir>
#include <QTextStream>
#include <QMessageBox>
#include "mydebug.h"

BPPEngine::BPPEngine(const QString &bppPath, QTextEdit *aOutputWindow, int msec1, int msec2, QObject *parent) :
    Engine(aOutputWindow, msec1, msec2, parent)
{
    setPath(bppPath);
    setDefaultData();
}

bool BPPEngine::setPath(const QString& bppPath)
{
#ifdef Q_OS_WIN32
    Engine::setPath( bppPath + "/bpp.exe" );
#else
    Engine::setPath( bppPath + "/bpp" );
#endif

    setVersion( (isStatusOk())? ::getVersionByPath(bppPath) : QString("") );

    return isStatusOk();
}

void BPPEngine::setDefaultData()
{
    nEdit_seed = -1;
    nComb_noisy = 3;
    strEdit_seqfile = "";
    strEdit_Imapfile = "";
    strEdit_outfile_Name = "out.txt";
    strEdit_mcmcfile_Name = "mcmc.txt";
    strEdit_Output_Location = "";
    nComb_speciesdelimitation = 1;
    nComb_RJalgorithm = 0;
    nEdit_RJfinetune[0] = 2;
    nEdit_RJfinetune[1] = 0;
    nComb_speciestree = 1;
    nComb_speciesmodelprior = 1;
    nEdit_nspecies = 5;
    cList_species.clear();
    cList_species.append(Species("A", 3));
    cList_species.append(Species("B", 1));
    cList_species.append(Species("C", 2));
    cList_species.append(Species("D", 1));
    cList_species.append(Species("E", 1));
    strEdit_tree = "(((A, B), C), (D, E))";
    nComb_usedata = 1;
    nEdit_nloci = 5;
    bButt_cleandata = true;
    nEdit_thetaprior[0] = 2;
    nEdit_thetaprior[1] = 2000;
    nEdit_tauprior[0] = 2;
    nEdit_tauprior[1] = 20000;
    nEdit_tauprior[2] = 1;
    nComb_locusrate = 0;
    nEdit_a_locusrate = 2.0;
    strEdit_locusratefile = "";
    nComb_heredity = 0;
    nEdit_a_heredity = 4;
    nEdit_b_heredity = 4;
    strEdit_heredityfile = "";
//    nComb_sequenceerror;
    bButt_autoAdjustFinetune = true;
    nEdit_finetune[0] = 0.01;
    nEdit_finetune[1] = 0.01;
    nEdit_finetune[2] = 0.01;
    nEdit_finetune[3] = 0.01;
    nEdit_finetune[4] = 0.01;
    nEdit_finetune[5] = 0.01;
    nEdit_finetune[6] = 0.01;
    bButt_print = true;
    bButt_printlocusrate = false;
    bButt_printheredity = false;
    bButt_printGenetree = false;
    nEdit_burnin = 20000;
    nEdit_sampfreq = 5;
    nEdit_nsample = 100000;
}

bool BPPEngine::readFile(const QString& fileName)
{
    CStyleBPPCtlLoader* ctlLoader = new CStyleBPPCtlLoader(this);
    QFile ctlFile(fileName);

    ctlLoader->seed = -1;
    ctlLoader->noisy = 3;
    ctlLoader->sptree.speciesdelimitation = 2;
    ctlLoader->mcmc.RJalgorithm = 0;
    ctlLoader->mcmc.RJfinetune[0] = 2;
    ctlLoader->mcmc.RJfinetune[1] = 0;
    ctlLoader->sptree.speciestree = 1;
    ctlLoader->sptree.SpeciesModelPrior = 1;
    ctlLoader->sptree.nspecies = 0;
    ctlLoader->mcmc.usedata = 1;
    ctlLoader->data.ngene = 5;
    ctlLoader->com.cleandata = char(1);
    ctlLoader->data.theta_prior[0] = 2;
    ctlLoader->data.theta_prior[1] = 2000;
    ctlLoader->data.tau_prior[0] = 2;
    ctlLoader->data.tau_prior[1] = 20000;
    ctlLoader->data.tau_prior[2] = 1;
    ctlLoader->data.est_locusrate = 0;
    ctlLoader->data.a_locusrate = 2.0;
    ctlLoader->data.est_heredity = 0;
    ctlLoader->data.a_heredity = 4;
    ctlLoader->data.b_heredity = 4;
//    nComb_sequenceerror;
    ctlLoader->mcmc.resetFinetune = 1;
    ctlLoader->mcmc.finetune[0] = 0.01;
    ctlLoader->mcmc.finetune[1] = 0.01;
    ctlLoader->mcmc.finetune[2] = 0.01;
    ctlLoader->mcmc.finetune[3] = 0.01;
    ctlLoader->mcmc.finetune[4] = 0.01;
    ctlLoader->mcmc.finetune[5] = 0.01;
    ctlLoader->mcmc.finetune[6] = 0.01;
    ctlLoader->mcmc.print = true;
    ctlLoader->mcmc.printlocusrate = false;
    ctlLoader->mcmc.printheredity = false;
    ctlLoader->mcmc.printGenetree = false;
    ctlLoader->mcmc.burnin = 20000;
    ctlLoader->mcmc.sampfreq = 5;
    ctlLoader->mcmc.nsample = 100000;

    if( ctlLoader->GetOptions(fileName.toLocal8Bit().data()) != 0 || !ctlFile.open(QIODevice::ReadOnly) ) {
        delete ctlLoader;

        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Critical);
        msgBox.setWindowTitle(tr("Error !"));
        msgBox.setText(tr("Failed to read file:"));
        msgBox.setInformativeText(QDir::toNativeSeparators(fileName));
        msgBox.exec();
        return false;
    }

    QTextStream ctlIn(&ctlFile);
    mFileContent = ctlIn.readAll();

    nEdit_seed = ctlLoader->seed;
    nComb_noisy = ctlLoader->noisy;
    strEdit_seqfile = ctlLoader->seqfile.section('*', 0, 0).trimmed();
    strEdit_Imapfile = ctlLoader->Imapfile.section('*', 0, 0).trimmed();
    QString outfile = ctlLoader->outfile.section('*', 0, 0).trimmed();
    strEdit_outfile_Name = QFileInfo(outfile).fileName();
    strEdit_Output_Location = QDir::toNativeSeparators(QFileInfo(outfile).path());
    if(strEdit_Output_Location == ".") strEdit_Output_Location = QDir::toNativeSeparators("./");
    if(strEdit_Output_Location == "..") strEdit_Output_Location = QDir::toNativeSeparators("../");
    QString mcmcfile = ctlLoader->mcmcfile.section('*', 0, 0).trimmed();
    strEdit_mcmcfile_Name = QFileInfo(mcmcfile).fileName();
    nComb_speciesdelimitation = ctlLoader->sptree.speciesdelimitation;
    nComb_RJalgorithm = ctlLoader->mcmc.RJalgorithm;
    nEdit_RJfinetune[0] = ctlLoader->mcmc.RJfinetune[0];
    nEdit_RJfinetune[1] = ctlLoader->mcmc.RJfinetune[1];
    nComb_speciestree = ctlLoader->sptree.speciestree;
    nComb_speciesmodelprior = ctlLoader->sptree.SpeciesModelPrior;
    nEdit_nspecies = ctlLoader->sptree.nspecies;
    cList_species.clear();
    for(int i=0; i<nEdit_nspecies; i++)
        cList_species.append( Species(ctlLoader->sptree.nodes[i].name, ctlLoader->sptree.nseqsp[i]) );
    strEdit_tree = ctlLoader->treeString.section('*', 0, 0).trimmed();
    while(strEdit_tree.endsWith(';')) strEdit_tree.remove(-1, 1);
    nComb_usedata = ctlLoader->mcmc.usedata;
    nEdit_nloci = ctlLoader->data.ngene;
    bButt_cleandata = ctlLoader->com.cleandata;
    nEdit_thetaprior[0] = ctlLoader->data.theta_prior[0];
    nEdit_thetaprior[1] = ctlLoader->data.theta_prior[1];
    nEdit_tauprior[0] = ctlLoader->data.tau_prior[0];
    nEdit_tauprior[1] = ctlLoader->data.tau_prior[1];
    nEdit_tauprior[2] = ctlLoader->data.tau_prior[2];
    nComb_locusrate = ctlLoader->data.est_locusrate;
    nEdit_a_locusrate = ctlLoader->data.a_locusrate;
    strEdit_locusratefile = ctlLoader->locusratefile.section('*', 0, 0).trimmed();
    nComb_heredity = ctlLoader->data.est_heredity;
    nEdit_a_heredity = ctlLoader->data.a_heredity;
    nEdit_b_heredity = ctlLoader->data.b_heredity;
    strEdit_heredityfile = ctlLoader->heredityfile.section('*', 0, 0).trimmed();
//    nComb_sequenceerror;
    bButt_autoAdjustFinetune = ctlLoader->mcmc.resetFinetune;
    nEdit_finetune[0] = ctlLoader->mcmc.finetune[0];
    nEdit_finetune[1] = ctlLoader->mcmc.finetune[1];
    nEdit_finetune[2] = ctlLoader->mcmc.finetune[2];
    nEdit_finetune[3] = ctlLoader->mcmc.finetune[3];
    nEdit_finetune[4] = ctlLoader->mcmc.finetune[4];
    nEdit_finetune[5] = ctlLoader->mcmc.finetune[5];
    nEdit_finetune[6] = ctlLoader->mcmc.finetune[6];
    bButt_print = ctlLoader->mcmc.print;
    bButt_printlocusrate = ctlLoader->mcmc.printlocusrate;
    bButt_printheredity = ctlLoader->mcmc.printheredity;
    bButt_printGenetree = ctlLoader->mcmc.printGenetree;
    nEdit_burnin = ctlLoader->mcmc.burnin;
    nEdit_sampfreq = ctlLoader->mcmc.sampfreq;
    nEdit_nsample = ctlLoader->mcmc.nsample;

    delete ctlLoader;

    return true;
}

bool BPPEngine::writeFile(const QString& fileName)
{
    refreshFileContent();

    QFile ctlFile(fileName);
    if( !ctlFile.open(QIODevice::WriteOnly) ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Critical);
        msgBox.setWindowTitle(tr("Error !"));
        msgBox.setText(tr("Failed to write to file:"));
        msgBox.setInformativeText(QDir::toNativeSeparators(fileName));
        msgBox.exec();
        return false;
    }
    QTextStream ctlOut(&ctlFile);

    ctlOut << mFileContent;
    ctlOut.flush();

    return true;
}

bool BPPEngine::createTmpCtlFile(const QString& fileName)
{
    QFile ctlFile(fileName);
    if( !ctlFile.open(QIODevice::WriteOnly) ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Critical);
        msgBox.setWindowTitle(tr("Error !"));
        msgBox.setText(tr("Failed to create or open file:"));
        msgBox.setInformativeText(QDir::toNativeSeparators(fileName));
        msgBox.exec();
        return false;
    }
    QTextStream ctlOut(&ctlFile);

    ctlOut << "          seed = " << nEdit_seed << "\n\n";

    if(nComb_noisy < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Noisy option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>noisy</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "         noisy = " << nComb_noisy << "\n\n";

    strEdit_seqfile = strEdit_seqfile.trimmed();
    if( strEdit_seqfile.isEmpty() ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Sequence file's path can not be empty."));
        msgBox.setInformativeText(tr("Please enter <i>seqfile</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_seqfile.length() >= MAX_PATH ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Full name of sequence file is too long."));
        msgBox.setInformativeText(tr("Please check <i>seqfile</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_seqfile.contains(' ') ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("There should be no space characters in sequence file's path."));
        msgBox.setInformativeText(tr("Please check <i>seqfile</i>."));
        msgBox.exec();
        return false;
    }
    if( QDir::isAbsolutePath(strEdit_seqfile) || mFileName.isEmpty() ) {
        if( !QDir::isAbsolutePath(strEdit_seqfile) || !QFileInfo(strEdit_seqfile).isFile() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Failed to open sequence file."));
            msgBox.setInformativeText(tr("Please check <i>seqfile</i>."));
            msgBox.exec();
            return false;
        }
        ctlOut << "       seqfile = " << strEdit_seqfile << "\n";
    }
    else {
        QString seqfilePath = QDir::cleanPath( QFileInfo(mFileName).absoluteDir().absoluteFilePath(strEdit_seqfile) );
        if( !QFileInfo(seqfilePath).isFile() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Failed to open sequence file."));
            msgBox.setInformativeText(tr("Please check <i>seqfile</i>."));
            msgBox.exec();
            return false;
        }
        if( seqfilePath.length() >= MAX_PATH ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Absolute path of sequence file is too long."));
            msgBox.setInformativeText(QDir::toNativeSeparators(seqfilePath));
            msgBox.exec();
            return false;
        }
        if( seqfilePath.contains(' ') ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Absolute path of sequence file can not have space characters."));
            msgBox.setInformativeText(QDir::toNativeSeparators(seqfilePath));
            msgBox.exec();
            return false;
        }
        ctlOut << "       seqfile = " << seqfilePath << "\n";
    }

    if(nEdit_nspecies > 1) {
        strEdit_Imapfile = strEdit_Imapfile.trimmed();
        if( strEdit_Imapfile.isEmpty() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Individual map file's path can not be empty."));
            msgBox.setInformativeText(tr("Please enter <i>Imapfile</i>."));
            msgBox.exec();
            return false;
        }
        if( strEdit_Imapfile.length() >= MAX_PATH ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Full name of individual map file is too long."));
            msgBox.setInformativeText(tr("Please check <i>Imapfile</i>."));
            msgBox.exec();
            return false;
        }
        if( strEdit_Imapfile.contains(' ') ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("There should be no space characters in individual map file's path."));
            msgBox.setInformativeText(tr("Please check <i>Imapfile</i>."));
            msgBox.exec();
            return false;
        }
        if( QDir::isAbsolutePath(strEdit_Imapfile) || mFileName.isEmpty() ) {
            if( !QDir::isAbsolutePath(strEdit_Imapfile) || !QFileInfo(strEdit_Imapfile).isFile() ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Failed to open individual map file."));
                msgBox.setInformativeText(tr("Please check <i>Imapfile</i>."));
                msgBox.exec();
                return false;
            }
            ctlOut << "      Imapfile = " << strEdit_Imapfile << "\n";
        }
        else {
            QString ImapfilePath = QDir::cleanPath( QFileInfo(mFileName).absoluteDir().absoluteFilePath(strEdit_Imapfile) );
            if( !QFileInfo(ImapfilePath).isFile() ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Failed to open individual map file."));
                msgBox.setInformativeText(tr("Please check <i>Imapfile</i>."));
                msgBox.exec();
                return false;
            }
            if( ImapfilePath.length() >= MAX_PATH ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Absolute path of individual map file is too long."));
                msgBox.setInformativeText(QDir::toNativeSeparators(ImapfilePath));
                msgBox.exec();
                return false;
            }
            if( ImapfilePath.contains(' ') ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Absolute path of individual map file can not have space characters."));
                msgBox.setInformativeText(QDir::toNativeSeparators(ImapfilePath));
                msgBox.exec();
                return false;
            }
            ctlOut << "      Imapfile = " << ImapfilePath << "\n";
        }
    }

    strEdit_outfile_Name = strEdit_outfile_Name.trimmed();
    if( strEdit_outfile_Name.isEmpty() ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("outfile's name can not be empty."));
        msgBox.setInformativeText(tr("Please enter <i>outfile name</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_outfile_Name.length() >= MAX_PATH ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Name of outfile is too long."));
        msgBox.setInformativeText(tr("Please check <i>outfile name</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_outfile_Name.contains(' ') ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("There should be no space characters in outfile's name."));
        msgBox.setInformativeText(tr("Please check <i>outfile name</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "       outfile = " << strEdit_outfile_Name << "\n";

    strEdit_mcmcfile_Name = strEdit_mcmcfile_Name.trimmed();
    if( strEdit_mcmcfile_Name.isEmpty() ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("mcmcfile's name can not be empty."));
        msgBox.setInformativeText(tr("Please enter <i>mcmcfile name</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_mcmcfile_Name.length() >= MAX_PATH ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Name of mcmcfile is too long."));
        msgBox.setInformativeText(tr("Please check <i>mcmcfile name</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_mcmcfile_Name.contains(' ') ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("There should be no space characters in mcmcfile's name."));
        msgBox.setInformativeText(tr("Please check <i>mcmcfile name</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "      mcmcfile = " << strEdit_mcmcfile_Name << "\n\n";

    if(nComb_speciesdelimitation < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Speciesdelimitation option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>speciesdelimitation</i>."));
        msgBox.exec();
        return false;
    }
    if(nComb_RJalgorithm < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("RJalgorithm option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>RJalgorithm</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "  speciesdelimitation = " << nComb_speciesdelimitation << ' ' << nComb_RJalgorithm << ' ' << nEdit_RJfinetune[0];
    if(nComb_RJalgorithm == 1)
        ctlOut << ' ' << nEdit_RJfinetune[1];
    ctlOut << "\n";
    if(nComb_speciestree < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Speciestree option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>speciestree</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "          speciestree = " << nComb_speciestree << "\n\n";

    if(nComb_speciesmodelprior < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Speciesmodelprior option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>speciesmodelprior</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "speciesmodelprior = " << nComb_speciesmodelprior << "\n\n";

    ctlOut << "  species&tree = " << nEdit_nspecies;
    int* spaceCount = new int[nEdit_nspecies+1];
    spaceCount[0] = 2;
    if(nEdit_nspecies>=10 && nEdit_nspecies<=99) spaceCount[0]++;
    for(int i=0; i<nEdit_nspecies; i++) {
        cList_species[i].name = cList_species[i].name.trimmed();
        if( cList_species[i].name.isEmpty() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Species name can not be empty."));
            msgBox.setInformativeText(tr("Please enter <i>species name</i>."));
            msgBox.exec();
            return false;
        }
        ctlOut << "  " << cList_species[i].name;
        spaceCount[i+1] = 2 + cList_species[i].name.size() - 1;
    }
    ctlOut << "\n                  ";
    for(int i=0; i<nEdit_nspecies; i++)
        ctlOut << QString(spaceCount[i], ' ') << cList_species[i].nseqsp;
    if(nEdit_nspecies > 1) {
        strEdit_tree = strEdit_tree.trimmed();
        if( strEdit_tree.isEmpty() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Species tree can not be empty."));
            msgBox.setInformativeText(tr("Please enter <i>species tree</i>."));
            msgBox.exec();
            return false;
        }
        ctlOut << "\n                 " << strEdit_tree << ';';
    }
    ctlOut << "\n\n";
    delete[]spaceCount;
    spaceCount = NULL;

    if(nComb_usedata < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Usedata option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>usedata</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "       usedata = " << nComb_usedata << "\n";
    ctlOut << "         nloci = " << nEdit_nloci << "\n\n";

    ctlOut << "     cleandata = " << bButt_cleandata << "\n\n";

    ctlOut << "    thetaprior = " << nEdit_thetaprior[0] << ' ' << nEdit_thetaprior[1] << "\n";
    ctlOut << "      tauprior = " << nEdit_tauprior[0] << ' ' << nEdit_tauprior[1] << ' ' << nEdit_tauprior[2] << "\n\n";

    if(nComb_locusrate < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Locusrate option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>locusrate</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "     locusrate = " << nComb_locusrate << ' ' << nEdit_a_locusrate;
    if(nComb_locusrate == 2) {
        strEdit_locusratefile = strEdit_locusratefile.trimmed();
        if( strEdit_locusratefile.isEmpty() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Locus rate file's path can not be empty."));
            msgBox.setInformativeText(tr("Please enter <i>locus rate file</i>."));
            msgBox.exec();
            return false;
        }
        if( strEdit_locusratefile.length() >= MAX_PATH ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Full name of locus rate file is too long."));
            msgBox.setInformativeText(tr("Please check <i>locus rate file</i>."));
            msgBox.exec();
            return false;
        }
        if( strEdit_locusratefile.contains(' ') ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("There should be no space characters in locus rate file's path."));
            msgBox.setInformativeText(tr("Please check <i>locus rate file</i>."));
            msgBox.exec();
            return false;
        }
        if( QDir::isAbsolutePath(strEdit_locusratefile) || mFileName.isEmpty() ) {
            if( !QDir::isAbsolutePath(strEdit_locusratefile) || !QFileInfo(strEdit_locusratefile).isFile() ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Failed to open locus rate file."));
                msgBox.setInformativeText(tr("Please check <i>locus rate file</i>."));
                msgBox.exec();
                return false;
            }
            ctlOut << ' ' << strEdit_locusratefile;
        }
        else {
            QString locusratefilePath = QDir::cleanPath( QFileInfo(mFileName).absoluteDir().absoluteFilePath(strEdit_locusratefile) );
            if( !QFileInfo(locusratefilePath).isFile() ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Failed to open locus rate file."));
                msgBox.setInformativeText(tr("Please check <i>locus rate file</i>."));
                msgBox.exec();
                return false;
            }
            if( locusratefilePath.length() >= MAX_PATH ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Absolute path of locus rate file is too long."));
                msgBox.setInformativeText(QDir::toNativeSeparators(locusratefilePath));
                msgBox.exec();
                return false;
            }
            if( locusratefilePath.contains(' ') ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Absolute path of locus rate file can not have space characters."));
                msgBox.setInformativeText(QDir::toNativeSeparators(locusratefilePath));
                msgBox.exec();
                return false;
            }
            ctlOut << ' ' << locusratefilePath;
        }
    }
    ctlOut << "\n";
    if(nComb_heredity < 0) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Heredity option can not be empty."));
        msgBox.setInformativeText(tr("Please select <i>heredity</i>."));
        msgBox.exec();
        return false;
    }
    ctlOut << "      heredity = " << nComb_heredity << ' ' << nEdit_a_heredity << ' ' << nEdit_b_heredity;
    if(nComb_heredity == 2) {
        strEdit_heredityfile = strEdit_heredityfile.trimmed();
        if( strEdit_heredityfile.isEmpty() ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Heredity file's path can not be empty."));
            msgBox.setInformativeText(tr("Please enter <i>heredity file</i>."));
            msgBox.exec();
            return false;
        }
        if( strEdit_heredityfile.length() >= MAX_PATH ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("Full name of heredity file is too long."));
            msgBox.setInformativeText(tr("Please check <i>heredity file</i>."));
            msgBox.exec();
            return false;
        }
        if( strEdit_heredityfile.contains(' ') ) {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Information);
            msgBox.setWindowTitle(tr("Warning"));
            msgBox.setText(tr("There should be no space characters in heredity file's path."));
            msgBox.setInformativeText(tr("Please check <i>heredity file</i>."));
            msgBox.exec();
            return false;
        }
        if( QDir::isAbsolutePath(strEdit_heredityfile) || mFileName.isEmpty() ) {
            if( !QDir::isAbsolutePath(strEdit_heredityfile) || !QFileInfo(strEdit_heredityfile).isFile() ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Failed to open heredity file."));
                msgBox.setInformativeText(tr("Please check <i>heredity file</i>."));
                msgBox.exec();
                return false;
            }
            ctlOut << ' ' << strEdit_heredityfile;
        }
        else {
            QString heredityfilePath = QDir::cleanPath( QFileInfo(mFileName).absoluteDir().absoluteFilePath(strEdit_heredityfile) );
            if( !QFileInfo(heredityfilePath).isFile() ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Failed to open heredity file."));
                msgBox.setInformativeText(tr("Please check <i>heredity file</i>."));
                msgBox.exec();
                return false;
            }
            if( heredityfilePath.length() >= MAX_PATH ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Absolute path of heredity file is too long."));
                msgBox.setInformativeText(QDir::toNativeSeparators(heredityfilePath));
                msgBox.exec();
                return false;
            }
            if( heredityfilePath.contains(' ') ) {
                QMessageBox msgBox;
                msgBox.setIcon(QMessageBox::Information);
                msgBox.setWindowTitle(tr("Warning"));
                msgBox.setText(tr("Absolute path of heredity file can not have space characters."));
                msgBox.setInformativeText(QDir::toNativeSeparators(heredityfilePath));
                msgBox.exec();
                return false;
            }
            ctlOut << ' ' << heredityfilePath;
        }
    }
    ctlOut << "\n\n";

    ctlOut << "      finetune = " << char('0' + bButt_autoAdjustFinetune) << ": " << nEdit_finetune[0] << ' ' << nEdit_finetune[1] << ' ' << nEdit_finetune[2] << ' ' << nEdit_finetune[3] << ' ' << nEdit_finetune[4] << ' ' << nEdit_finetune[5] << ' ' << nEdit_finetune[6] << ' ' << "\n\n";

    ctlOut << "         print = " << char('0' + bButt_print) << ' ' << char('0' + bButt_printlocusrate) << ' ' << char('0' + bButt_printheredity) << ' ' << char('0' + bButt_printGenetree) << "\n";
    ctlOut << "        burnin = " << nEdit_burnin << "\n";
    ctlOut << "      sampfreq = " << nEdit_sampfreq << "\n";
    ctlOut << "       nsample = " << nEdit_nsample << "\n\n";

    ctlOut.flush();

    return true;
}

void BPPEngine::refreshFileContent()
{
    if(!modified && !mFileName.isEmpty()) return;

    QString tmpContent = mFileContent;
    mFileContent.clear();

    QTextStream in(&tmpContent, QIODevice::ReadOnly);
    QTextStream out(&mFileContent, QIODevice::WriteOnly);

    QStringList optionList;
    optionList << "seed"
               << "noisy"
               << "seqfile"
               << "Imapfile"
               << "outfile"
               << "mcmcfile"
               << "speciesdelimitation"
               << "speciestree"
               << "speciesmodelprior"
               << "species&tree"
               << "usedata"
               << "nloci"
               << "cleandata"
               << "thetaprior"
               << "tauprior"
               << "locusrate"
               << "heredity"
//               << "sequenceerror"
               << "finetune"
               << "print"
               << "burnin"
               << "sampfreq"
               << "nsample";

    int size= optionList.size();
    bool* used = new bool[size];
    for(int i=0; i<size; i++) used[i] = false;

    while(!in.atEnd()) {
        QString line = in.readLine();
        QString optionString = line.section('#', 0, 0).section('*', 0, 0).section('=', 0, 0);
        int pos1 = line.indexOf('*');
        int pos2 = line.indexOf('#');
        QChar ch = (pos1 < pos2)? '*' : '#';
        QString comment = ch + line.section(ch, 1);
        QString option = optionString.trimmed();

        if(option.isEmpty()) {
            out << line << "\n";
            continue;
        }

        int i=0;
        for(; i<size; i++) {
            if(option == optionList[i]) {
                if(!optionString.endsWith(' ')) optionString.append(' ');

                out << optionString << "= ";

                switch(i)
                {
                case 0:
                    out << nEdit_seed;
                    break;
                case 1:
                    if(nComb_noisy >= 0)
                        out << nComb_noisy;
                    break;
                case 2:
                    strEdit_seqfile = strEdit_seqfile.trimmed();
                    out << strEdit_seqfile;
                    break;
                case 3:
                    strEdit_Imapfile = strEdit_Imapfile.trimmed();
                    out << strEdit_Imapfile;
                    break;
                case 4:
                    strEdit_Output_Location = strEdit_Output_Location.trimmed();
                    strEdit_outfile_Name = strEdit_outfile_Name.trimmed();
                    if(strEdit_Output_Location.isEmpty())
                        out << strEdit_outfile_Name;
                    else if(strEdit_Output_Location == "/" || strEdit_Output_Location == "\\")
                        out << QDir::fromNativeSeparators(QDir::cleanPath( "/" + strEdit_outfile_Name ));
                    else
                        out << QDir::fromNativeSeparators(QDir::cleanPath( strEdit_Output_Location + "/" + strEdit_outfile_Name ));
                    break;
                case 5:
                    strEdit_Output_Location = strEdit_Output_Location.trimmed();
                    strEdit_mcmcfile_Name = strEdit_mcmcfile_Name.trimmed();
                    if(strEdit_Output_Location.isEmpty())
                        out << strEdit_mcmcfile_Name;
                    else if(strEdit_Output_Location == "/" || strEdit_Output_Location == "\\")
                        out << QDir::fromNativeSeparators(QDir::cleanPath( "/" + strEdit_mcmcfile_Name ));
                    else
                        out << QDir::fromNativeSeparators(QDir::cleanPath( strEdit_Output_Location + "/" + strEdit_mcmcfile_Name ));
                    break;
                case 6:
                    if(nComb_speciesdelimitation >= 0 && nComb_RJalgorithm >= 0) {
                        out << nComb_speciesdelimitation << ' ' << nComb_RJalgorithm << ' ' << nEdit_RJfinetune[0];
                        if(nComb_RJalgorithm == 1)
                            out << ' ' << nEdit_RJfinetune[1];
                    }
                    break;
                case 7:
                    if(nComb_speciestree >= 0)
                        out << nComb_speciestree;
                    break;
                case 8:
                    if(nComb_speciesmodelprior >= 0)
                        out << nComb_speciesmodelprior;
                    break;
                case 9:
                    {
                    out << nEdit_nspecies;
                    int* spaceCount = new int[nEdit_nspecies+1];
                    spaceCount[0] = 2;
                    if(nEdit_nspecies>=10 && nEdit_nspecies<=99) spaceCount[0]++;
                    for(int i=0; i<nEdit_nspecies; i++) {
                        cList_species[i].name = cList_species[i].name.trimmed();
                        out << "  " << cList_species[i].name;
                        spaceCount[i+1] = 2 + cList_species[i].name.size() - 1;
                    }

                    if(comment.size() > 1)
                        out << "   " << comment << "\n";
                    else
                        out << "\n";
                    comment.clear();

                    out << QString(optionString.size()+3, ' ');
                    for(int i=0; i<nEdit_nspecies; i++)
                        out << QString(spaceCount[i], ' ') << cList_species[i].nseqsp;

                    QString line_local = in.readLine();
                    int pos1_local = line_local.indexOf('*');
                    int pos2_local = line_local.indexOf('#');
                    QChar ch_local = (pos1_local < pos2_local)? '*' : '#';
                    QString comment_local = ch_local + line_local.section(ch_local, 1);
                    if(comment_local.size() > 1)
                        out << "   " << comment_local << "\n";
                    else
                        out << "\n";

                    strEdit_tree = strEdit_tree.trimmed();
                    out << QString(optionString.size()+2, ' ') << strEdit_tree << ';';

                    line_local = in.readLine();
                    pos1_local = line_local.indexOf('*');
                    pos2_local = line_local.indexOf('#');
                    ch_local = (pos1_local < pos2_local)? '*' : '#';
                    comment_local = ch_local + line_local.section(ch_local, 1);
                    if(comment_local.size() > 1)
                        out << "   " << comment_local;

                    delete[]spaceCount;
                    spaceCount = NULL;
                    }
                    break;
                case 10:
                    if(nComb_usedata >= 0)
                        out << nComb_usedata;
                    break;
                case 11:
                    out << nEdit_nloci;
                    break;
                case 12:
                    out << char('0' + bButt_cleandata);
                    break;
                case 13:
                    out << nEdit_thetaprior[0] << ' ' << nEdit_thetaprior[1];
                    break;
                case 14:
                    out << nEdit_tauprior[0] << ' ' << nEdit_tauprior[1] << ' ' << nEdit_tauprior[2];
                    break;
                case 15:
                    if(nComb_locusrate >= 0) {
                        out << nComb_locusrate << ' ' << nEdit_a_locusrate;
                        if(nComb_locusrate == 2) {
                            strEdit_locusratefile = strEdit_locusratefile.trimmed();
                            out << ' ' << strEdit_locusratefile;
                        }
                    }
                    break;
                case 16:
                    if(nComb_heredity >= 0) {
                        out << nComb_heredity << ' ' << nEdit_a_heredity << ' ' << nEdit_b_heredity;
                        if(nComb_heredity == 2) {
                            strEdit_heredityfile = strEdit_heredityfile.trimmed();
                            out << ' ' << strEdit_heredityfile;
                        }
                    }
                    break;
                case 17:
                    out << char('0' + bButt_autoAdjustFinetune) << ": " << nEdit_finetune[0] << ' ' << nEdit_finetune[1] << ' ' << nEdit_finetune[2] << ' ' << nEdit_finetune[3] << ' ' << nEdit_finetune[4] << ' ' << nEdit_finetune[5] << ' ' << nEdit_finetune[6] << ' ';
                    break;
                case 18:
                    out << char('0' + bButt_print) << ' ' << char('0' + bButt_printlocusrate) << ' ' << char('0' + bButt_printheredity) << ' ' << char('0' + bButt_printGenetree);
                    break;
                case 19:
                    out << nEdit_burnin;
                    break;
                case 20:
                    out << nEdit_sampfreq;
                    break;
                case 21:
                    out << nEdit_nsample;
                    break;
                default:;
                }

                if(comment.size() > 1)
                    out << "   " << comment << "\n";
                else
                    out << "\n";

                used[i] = true;
                break;
            }
        }

        if(i == size) {
            out << line << "\n";
        }
    }

    for(int i=0; i<size; i++) {
        if(!used[i]) {
            switch(i)
            {
            case 0:
                out << "          seed = " << nEdit_seed << "\n\n";
                break;
            case 1:
                if(nComb_noisy >= 0)
                    out << "         noisy = " << nComb_noisy << "\n\n";
                break;
            case 2:
                strEdit_seqfile = strEdit_seqfile.trimmed();
                out << "       seqfile = " << strEdit_seqfile << "\n";
                break;
            case 3:
                if(nEdit_nspecies > 1) {
                    strEdit_Imapfile = strEdit_Imapfile.trimmed();
                    out << "      Imapfile = " << strEdit_Imapfile << "\n";
                }
                break;
            case 4:
                strEdit_Output_Location = strEdit_Output_Location.trimmed();
                strEdit_outfile_Name = strEdit_outfile_Name.trimmed();
                if(strEdit_Output_Location.isEmpty())
                    out << "       outfile = " << strEdit_outfile_Name << "\n";
                else if(strEdit_Output_Location == "/" || strEdit_Output_Location == "\\")
                    out << "       outfile = " << QDir::fromNativeSeparators(QDir::cleanPath( "/" + strEdit_outfile_Name )) << "\n";
                else
                    out << "       outfile = " << QDir::fromNativeSeparators(QDir::cleanPath( strEdit_Output_Location + "/" + strEdit_outfile_Name )) << "\n";
                break;
            case 5:
                strEdit_Output_Location = strEdit_Output_Location.trimmed();
                strEdit_mcmcfile_Name = strEdit_mcmcfile_Name.trimmed();
                if(strEdit_Output_Location.isEmpty())
                    out << "      mcmcfile = " << strEdit_mcmcfile_Name << "\n\n";
                else if(strEdit_Output_Location == "/" || strEdit_Output_Location == "\\")
                    out << "      mcmcfile = " << QDir::fromNativeSeparators(QDir::cleanPath( "/" + strEdit_mcmcfile_Name )) << "\n\n";
                else
                    out << "      mcmcfile = " << QDir::fromNativeSeparators(QDir::cleanPath( strEdit_Output_Location + "/" + strEdit_mcmcfile_Name )) << "\n\n";
                break;
            case 6:
                if(nComb_speciesdelimitation >= 0 && nComb_RJalgorithm >= 0) {
                    out << "  speciesdelimitation = " << nComb_speciesdelimitation << ' ' << nComb_RJalgorithm << ' ' << nEdit_RJfinetune[0];
                    if(nComb_RJalgorithm == 1)
                        out << ' ' << nEdit_RJfinetune[1];
                    out << "\n";
                }
                else
                    out << "  speciesdelimitation = \n";
                break;
            case 7:
                if(nComb_speciestree >= 0)
                    out << "speciestree = " << nComb_speciestree << "\n\n";
                else
                    out << "speciestree = \n\n";
                break;
            case 8:
                if(nComb_speciesmodelprior >= 0)
                    out << "speciesmodelprior = " << nComb_speciesmodelprior << "\n\n";
                else
                    out << "speciesmodelprior = \n\n";
                break;
            case 9:
                {
                out << "  species&tree = " << nEdit_nspecies;
                int* spaceCount = new int[nEdit_nspecies+1];
                spaceCount[0] = 2;
                if(nEdit_nspecies>=10 && nEdit_nspecies<=99) spaceCount[0]++;
                for(int i=0; i<nEdit_nspecies; i++) {
                    cList_species[i].name = cList_species[i].name.trimmed();
                    out << "  " << cList_species[i].name;
                    spaceCount[i+1] = 2 + cList_species[i].name.size() - 1;
                }
                out << "\n                  ";
                for(int i=0; i<nEdit_nspecies; i++)
                    out << QString(spaceCount[i], ' ') << cList_species[i].nseqsp;
                if(nEdit_nspecies > 1) {
                    strEdit_tree = strEdit_tree.trimmed();
                    out << "\n                 " << strEdit_tree << ';';
                }
                out << "\n\n";
                delete[]spaceCount;
                spaceCount = NULL;
                }
                break;
            case 10:
                if(nComb_usedata >= 0)
                    out << "       usedata = " << nComb_usedata << "\n";
                else
                    out << "       usedata = \n";
                break;
            case 11:
                out << "         nloci = " << nEdit_nloci << "\n\n";
                break;
            case 12:
                out << "     cleandata = " << char('0' + bButt_cleandata) << "\n\n";
                break;
            case 13:
                out << "    thetaprior = " << nEdit_thetaprior[0] << ' ' << nEdit_thetaprior[1] << "\n";
                break;
            case 14:
                out << "      tauprior = " << nEdit_tauprior[0] << ' ' << nEdit_tauprior[1] << ' ' << nEdit_tauprior[2] << "\n\n";
                break;
            case 15:
                if(nComb_locusrate >= 0) {
                    out << "     locusrate = " << nComb_locusrate << ' ' << nEdit_a_locusrate;
                    if(nComb_locusrate == 2) {
                        strEdit_locusratefile = strEdit_locusratefile.trimmed();
                        out << ' ' << strEdit_locusratefile;
                    }
                    out << "\n";
                }
                break;
            case 16:
                if(nComb_heredity >= 0) {
                    out << "      heredity = " << nComb_heredity << ' ' << nEdit_a_heredity << ' ' << nEdit_b_heredity;
                    if(nComb_heredity == 2) {
                        strEdit_heredityfile = strEdit_heredityfile.trimmed();
                        out << ' ' << strEdit_heredityfile;
                    }
                    out << "\n\n";
                }
                break;
            case 17:
                out << "      finetune = " << char('0' + bButt_autoAdjustFinetune) << ": " << nEdit_finetune[0] << ' ' << nEdit_finetune[1] << ' ' << nEdit_finetune[2] << ' ' << nEdit_finetune[3] << ' ' << nEdit_finetune[4] << ' ' << nEdit_finetune[5] << ' ' << nEdit_finetune[6] << ' ' << "\n\n";
                break;
            case 18:
                out << "         print = " << char('0' + bButt_print) << ' ' << char('0' + bButt_printlocusrate) << ' ' << char('0' + bButt_printheredity) << ' ' << char('0' + bButt_printGenetree) << "\n";
                break;
            case 19:
                out << "        burnin = " << nEdit_burnin << "\n";
                break;
            case 20:
                out << "      sampfreq = " << nEdit_sampfreq << "\n";
                break;
            case 21:
                out << "       nsample = " << nEdit_nsample << "\n\n";
                break;
            default:;
            }
        }
    }

    delete used;
    used = NULL;
}

bool BPPEngine::onNewDocument()
{
    if( !Engine::onNewDocument() )
        return false;

    mFileContent.clear();
    setDefaultData();

    return true;
}

bool BPPEngine::onCloseDocument()
{
    if( !Engine::onCloseDocument() )
        return false;

    mFileContent.clear();
    setDefaultData();

    return true;
}

bool BPPEngine::doOpen(const QString& fileName)
{
    return readFile(fileName);
}

bool BPPEngine::doSave(const QString& fileName)
{
    return writeFile(fileName);
}

bool BPPEngine::run()
{
    if( !Engine::run() )
        return false;

#ifdef STREAMLINED_OUTPUT
    streamlineFlag = 0;
#endif

    strEdit_Output_Location = strEdit_Output_Location.trimmed();
    if( strEdit_Output_Location.isEmpty() ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Output path can not be empty."));
        msgBox.setInformativeText(tr("Please enter <i>location</i>."));
        msgBox.exec();
        return false;
    }
    if( strEdit_Output_Location.length() >= MAX_PATH ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Output path is too long."));
        msgBox.setInformativeText(tr("Please check <i>location</i>."));
        msgBox.exec();
        return false;
    }

    QString workingDirectory = strEdit_Output_Location;

    if( !QDir::isAbsolutePath(workingDirectory) && !mFileName.isEmpty() ) {
        workingDirectory = QFileInfo(mFileName).absolutePath() + "/" + workingDirectory;
    }

    workingDirectory = QDir::cleanPath(workingDirectory);

    if( !QDir::isAbsolutePath(workingDirectory) || !QDir(workingDirectory).exists() ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Output path does not exist."));
        msgBox.setInformativeText(tr("Please check <i>location</i>."));
        msgBox.exec();
        return false;
    }

    if( workingDirectory.length() >= MAX_PATH ) {
        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Information);
        msgBox.setWindowTitle(tr("Warning"));
        msgBox.setText(tr("Absolute path of output is too long."));
        msgBox.setInformativeText(QDir::toNativeSeparators(workingDirectory));
        msgBox.exec();
        return false;
    }

    QString ctlFileName = (workingDirectory == "/")? "/bpp.ctl.tmp" : QDir::cleanPath( workingDirectory + "/bpp.ctl.tmp" );
    if( !createTmpCtlFile(ctlFileName) ) return false;

    if( !workingDirectory.isEmpty() ) //always true
        process()->setWorkingDirectory(workingDirectory);

    QStringList arguments;
    arguments << "bpp.ctl.tmp";
    if( compareVersion("3.1") >= 0 )
        arguments << "--stdout-no-buf";
    process()->start(path(), arguments);

    return true;
}

#ifdef STREAMLINED_OUTPUT
void BPPEngine::onReadyReadStandardOutput()
{
    QByteArray newData = process()->readAllStandardOutput();
    newData.prepend(buffer());
    int newDataSize = newData.size();

    int pos1 = newData.lastIndexOf('\r');
    int pos2 = newData.lastIndexOf('\n');
    int pos = (pos1 > pos2)? pos1 : pos2;

    buffer() = newData.right(newDataSize - (pos + 1));
    newData.remove(pos + 1, newDataSize);
    if(newData.isEmpty()) return;

    if(streamlineFlag == 2) {
#ifdef MYDEBUG
        fwrite(newData.data(), newData.size(), 1, mydebugfp);
        fflush(mydebugfp);
#endif
#if defined(Q_OS_WIN32)
        if(newData.endsWith("\r\n"))
            newData.chop(1);
#endif
        newData.chop(1);
        output(newData);
    }

    if(streamlineFlag == 0) {
        const char marker[] = "rior):";
        pos = newData.indexOf(marker);
        if(pos != -1) {
            //split newData = prefixData + newData
            int i = pos + sizeof(marker)/sizeof(char) - 1;
            for(; i<newData.size(); i++)
                if(newData[i] != '\r' && newData[i] != '\n') break;
            QByteArray prefixData = newData.left(i);
            newData.remove(0, i);
#ifdef MYDEBUG
            fwrite(prefixData.data(), prefixData.size(), 1, mydebugfp);
            fflush(mydebugfp);
#endif
#if defined(Q_OS_WIN32)
            if(prefixData.endsWith("\r\n"))
                prefixData.chop(1);
#endif
            prefixData.chop(1);
            output(prefixData);
            streamlineFlag = 1;
        }
        else {
#ifdef MYDEBUG
            fwrite(newData.data(), newData.size(), 1, mydebugfp);
            fflush(mydebugfp);
#endif
#if defined(Q_OS_WIN32)
            if(newData.endsWith("\r\n"))
                newData.chop(1);
#endif
            newData.chop(1);
            output(newData);
        }
    }

    if(streamlineFlag == 1) {
        QByteArray suffixData;
        pos = newData.indexOf("Time used");
        if(pos != -1) {
            //split newData = newData + suffixData
            newDataSize = newData.size();
            suffixData = newData.right(newDataSize - pos);
            newData.remove(pos, newDataSize);
            streamlineFlag = 2;
        }

        //streamline newData
#ifdef MYDEBUG
        fwrite(newData.data(), newData.size(), 1, mydebugfp);
        fflush(mydebugfp);
#endif
        char* string = newData.data();

#if defined(Q_OS_WIN32)
        for(char* index = string; *index != '\0'; index++) {
            if( *index == '\r' ) {
                if( *(index + 1) == '\n' ) {
                    *index = '\0';
                    output(string);
                    index++;
                }
                string = index + 1;
            }
        }
#elif defined(Q_OS_MAC) || defined(Q_OS_LINUX)
        for(char* index = string; *index != '\0'; index++) {
            if( *index == '\r' ) {
                string = index + 1;
            }
            else if( *index == '\n' ) {
                *index = '\0';
                output(string);
                string = index + 1;
            }
        }
#else
#warning Nothing to do on current platform
#endif

        if(streamlineFlag == 2) {
#ifdef MYDEBUG
            fwrite(suffixData.data(), suffixData.size(), 1, mydebugfp);
            fflush(mydebugfp);
#endif
#if defined(Q_OS_WIN32)
            if(suffixData.endsWith("\r\n"))
                suffixData.chop(1);
#endif
            suffixData.chop(1);
            output(suffixData);
        }
    }
}
#endif // STREAMLINED_OUTPUT
