//------------------------------------------------
// Copyright (C) 2012 Bo Xu <xuxbob@gmail.com>
//------------------------------------------------

#ifndef BPPENGINE_H
#define BPPENGINE_H

#include "engine.h"
#include "global.h"

#if defined(Q_OS_WIN32) || defined(Q_OS_MAC) || defined(Q_OS_LINUX)
#define STREAMLINED_OUTPUT
#endif

class BPPEngine : public Engine
{
    Q_OBJECT

public:
    explicit BPPEngine(const QString &bppPath, QTextEdit *aOutputWindow, int msec1, int msec2, QObject *parent = 0);

public:
    bool setPath(const QString& bppPath);
    int compareVersion(const QString& aVersion) const { return ::compareVersion(version(), aVersion); }

private:
    void setDefaultData();
    bool readFile(const QString& fileName);
    bool writeFile(const QString& fileName);
    bool createTmpCtlFile(const QString& fileName);

public:
    void refreshFileContent();
    bool onNewDocument();
    bool onCloseDocument();

protected:
    bool doOpen(const QString& fileName);
    bool doSave(const QString& fileName);

public slots:
    bool run();
#ifdef STREAMLINED_OUTPUT
    void onReadyReadStandardOutput();
#endif

public:
    struct Species {
        Species() {}
        Species(const QString& name, int nseqsp) : name(name), nseqsp(nseqsp) {}

        QString name;
        int nseqsp;
    };

public:
    int            nEdit_seed;
    int            nComb_noisy;
    QString        strEdit_seqfile;
    QString        strEdit_Imapfile;
    QString        strEdit_outfile_Name;
    QString        strEdit_mcmcfile_Name;
    QString        strEdit_Output_Location;
    int            nComb_speciesdelimitation;
    int            nComb_RJalgorithm;
    double         nEdit_RJfinetune[2];
    int            nComb_speciestree;
    int            nComb_speciesmodelprior;
    int            nEdit_nspecies;
    QList<Species> cList_species;
    QString        strEdit_tree;
    int            nComb_usedata;
    int            nEdit_nloci;
    bool           bButt_cleandata;
    double         nEdit_thetaprior[2];
    double         nEdit_tauprior[3];
    int            nComb_locusrate;
    double         nEdit_a_locusrate;
    QString        strEdit_locusratefile;
    int            nComb_heredity;
    double         nEdit_a_heredity;
    double         nEdit_b_heredity;
    QString        strEdit_heredityfile;
//    int            nComb_sequenceerror;
    bool           bButt_autoAdjustFinetune;
    double         nEdit_finetune[7];
    bool           bButt_print;
    bool           bButt_printlocusrate;
    bool           bButt_printheredity;
    bool           bButt_printGenetree;
    int            nEdit_burnin;
    int            nEdit_sampfreq;
    int            nEdit_nsample;

#ifdef STREAMLINED_OUTPUT
private:
    int streamlineFlag;
#endif
};

#endif // BPPENGINE_H
