#include "cstylebppctlloader.h"
#include "bppengine.h"
#include <cstdio>

int CStyleBPPCtlLoader::GetOptions(char *ctlf)
{
    int nopt=23, lline=4096, locfields[1000], iopt, i, is, ierror;
    char line[4096],*pline, opt[32]; const char *comment="*#"/*, *seqerrstr="0EF"*/;
    const char *optstr[] = {"seed", "noisy", "seqfile", "Imapfile", "outfile", "mcmcfile",
                            "speciesdelimitation", "speciestree", "speciesmodelprior", "species&tree", "usedata",
                            "nloci", "cleandata", "thetaprior", "tauprior", "locusrate", "heredity",
                            "sequenceerror", "finetune", "print", "burnin", "sampfreq", "nsample"};
    char name[LSPNAME];
    double t=1, *eps=mcmc.finetune;
    FILE  *fctl=fopen (ctlf, "r");

    sptree.SpeciesModelPrior = UniformRootedTree;
    data.lnpSpeciesModel = 0;
    mcmc.printGenetree=0;
    for(is=0; is<NSPECIES; is++)
       for(i=0; i<16; i++)
          data.a_seqerr[is][i] = data.e_seqerr[is][i] = 0;

    if (fctl) {
//       if (noisy) printf ("\nReading options from %s..\n", ctlf);
       for (;;) {
          if(fgets(line, lline, fctl) == NULL)
             break;
          if(line[0]=='/' && line[1]=='/')
             break;
          for (i=0,t=0,pline=line; i<lline&&line[i]; i++)
             if (isalnum(line[i]))  { t=1; break; }
             else if (strchr(comment,line[i])) break;
          if (t==0) continue;
          sscanf (line, "%s%*s%lf", opt, &t);
          if ((pline=strstr(line, "="))==NULL) { error2 ("option file.\nExpecting '=' "); return(-1); }

          for (iopt=0; iopt<nopt; iopt++) {
             if (strncmp(opt, optstr[iopt], 12)==0)  {
//                if (noisy>=9)
//                   printf ("\n%3d %15s | %-20s %6.2f", iopt+1,optstr[iopt],opt,t);
                switch (iopt) {
                   case ( 0): seed = (int)t;                      break;
                   case ( 1): sscanf(pline+1, "%d", &noisy);      break;
                   case ( 2):
                      seqfile = QString::fromLocal8Bit(pline+1);
//                      sscanf(pline+1, "%s", com.seqf);
                      break;
                   case ( 3):
                      Imapfile = QString::fromLocal8Bit(pline+1);
//                      sscanf(pline+1, "%s", com.Imapf);
                      break;
                   case ( 4):
                      outfile = QString::fromLocal8Bit(pline+1);
//                      sscanf(pline+1, "%s", com.outf);
                      break;
                   case ( 5):
                      mcmcfile = QString::fromLocal8Bit(pline+1);
//                      sscanf(pline+1, "%s", com.mcmcf);
                      break;
                   case ( 6):
                      sscanf(pline+1, "%d%d%lf%lf", &sptree.speciesdelimitation, &mcmc.RJalgorithm, &mcmc.RJfinetune[0], &mcmc.RJfinetune[1]);
                      if(sptree.speciesdelimitation==1) {
                         if(mcmc.RJalgorithm!=0 && mcmc.RJalgorithm!=1)  error2("RJalgorithm should be 0 or 1.");
                         if(mcmc.RJfinetune[0]<=0 || (mcmc.RJalgorithm==1 && mcmc.RJfinetune[1]<=0))
                            error2("RJfinetune <= 0?");
                      }
                      break;
                   case ( 7): sptree.speciestree = (int)t;  break;
                   case ( 8): sscanf(pline+1, "%d", &sptree.SpeciesModelPrior);   break;
                      break;
                   case ( 9):
                      if((sptree.nspecies=com.ns=(int)t)>NSPECIES)
                         error2("raise NSPECIES in bpp.c & recompile?");
                      ReadSpeciesTree(fctl, pline+1);
                      break;
                   case (10): mcmc.usedata=(int)t;   break;
                   case (11): data.ngene=(int)t;     break;
                   case (12): com.cleandata=(int)t;  break;
                   case (13): sscanf(pline+1, "%lf%lf", &data.theta_prior[0], &data.theta_prior[1]);
                      break;
                   case (14): sscanf(pline+1, "%lf%lf%lf", &data.tau_prior[0], &data.tau_prior[1], &data.tau_prior[2]);
                      break;
                   case (15):               /* locus rate */
                      sscanf(pline+1, "%d%lf", &data.est_locusrate, &data.a_locusrate);
                      if(data.est_locusrate==1) {
//                         printf("\nRates vary among loci, with ri/L ~ Dirichlet(%.2f)\n", data.a_locusrate);
                         if(data.a_locusrate<.001) error2("alpha very small?");
                      }
                      else if(data.est_locusrate==2) {
                         splitline(pline+1, locfields);
                         locusratefile = QString::fromLocal8Bit(pline+1+locfields[1]);
//                         sscanf(pline+1+locfields[1], "%s", com.locusratef); /* check this? */
                      }
                      break;
                   case (16):
                      sscanf(pline+1, "%d%lf%lf", &data.est_heredity, &data.a_heredity, &data.b_heredity);
                      if(data.est_heredity==1) {
//                         printf("\n\nheredity multiplier ~ gamma(%.2f, %.2f)\n", data.a_heredity, data.b_heredity);
                         if(data.a_heredity<.001 || data.b_heredity<.001) error2("poor prior?");
                      }
                      else if(data.est_heredity==2) {
                         splitline(pline+1, locfields);
                         heredityfile = QString::fromLocal8Bit(pline+1+locfields[1]);
//                         sscanf(pline+1+locfields[1], "%s", com.heredityf);
                      }
                      break;

                   case (17):               /* sequencing errors */
                      splitline(pline+1, locfields);
                      data.nseqerr = 0;
                      if(sscanf(pline+1, "%d", &data.nseqerr) != 1)
                         error2("error in the sequenceerror line of the control file.");
                      for(ierror=0; ierror<data.nseqerr; ierror++) {
                         fscanf(fctl, "%s", name);
                         for(is=0; is<sptree.nspecies; is++)
                            if( strcmp(name, sptree.nodes[is].name) == 0) break;
                         if(is==sptree.nspecies) error2("expecting a species name");
                         data.iseqerr[is] = 1;
                         for(i=0; i<16; i++)
                            fscanf(fctl, "%lf", &data.a_seqerr[is][i]);
//                         printf("\nalpha matrix for sequence errors in %s:", sptree.nodes[is].name);
//                         matout (F0, data.a_seqerr[is], 4, 4);
                      }
                      break;

                   case (18):
                      sscanf(pline+1,"%d:%lf%lf%lf%lf%lf%lf%lf", &mcmc.resetFinetune, eps,eps+1,eps+2,eps+3,eps+4,eps+5,eps+6);
                      break;
                   case (19):
                      sscanf(pline+1, "%d%d%d%d", &mcmc.print, &mcmc.printlocusrate, &mcmc.printheredity, &mcmc.printGenetree);
                      break;
                   case (20): mcmc.burnin=(int)t;    break;
                   case (21): mcmc.sampfreq=(int)t;  break;
                   case (22): mcmc.nsample=(int)t;   break;
                }
                break;
             }
          }
//          if (iopt==nopt)
//             { printf ("\noption %s in %s\n", opt, ctlf);  exit(-1); }
       }
       fclose(fctl);
    }
    else
       /*if (noisy)*/ { error2("\nno ctl file.."); return(-1); }

    if(sptree.SpeciesModelPrior==UserNodeProb && sptree.speciestree)
       error2("\nSpecies delimitation prior 2 does not work with the speciestree option");
                     
//    if(data.est_locusrate!=1) mcmc.printlocusrate=0;
//    if(data.est_heredity!=1)  mcmc.printheredity=0;
    return(0);
}

void CStyleBPPCtlLoader::error2(const char *message)
{
    QString msg(message);
    engine->output("\nError: " + msg + ".\n", QProcess::StandardError);
}

int CStyleBPPCtlLoader::splitline(char line[], int fields[])
{
/* This finds out how many fields there are in the line, and marks the starting
   positions of the fields.
   Fields are separated by spaces, and texts are allowed as well.
*/
    int lline=1000000, i, nfields=0, InSpace=1;
    char *p=line;

    for(i=0; i<lline && *p && *p!='\n'; i++,p++) {
       if (isspace(*p))
          InSpace=1;
       else  {
          if(InSpace) {
             InSpace=0;
             fields[nfields++]=i;
             if(nfields>MAXNFIELDS)
                /*puts*/error2("raise MAXNFIELDS?");
          }
       }
    }
    return(nfields);
}

int CStyleBPPCtlLoader::ReadSpeciesTree (FILE* fctl, char *currline)
{
/* It reads the species tree and initializes sptree.
   Note that (sptree.nseqsp[j] > 1) adds a population.
*/
    char line[16000]={'\0'};
    int i,/*j,*/ lline=16000, locfields[1000], maxns=0;

//    nodes = nodes_t;  /* species tree is read into the temp space nodes_t */
//    for(i=0; i<NS; i++) {  /* for both species tree & gene trees */
//       if (com.spname[i]) free(com.spname[i]);
//       com.spname[i] = (char*)malloc((LSPNAME+1)*sizeof(char));
//    }
    splitline (currline, locfields);
    for(i=0; i<sptree.nspecies; i++) {
       sscanf(currline+locfields[i+1], "%s", sptree.nodes[i].name);
//       sscanf(currline+locfields[i+1], "%s", com.spname[i]);
//       strcpy(sptree.nodes[i].name, com.spname[i]);
    }
    for(i=0,maxns=0; i<sptree.nspecies; i++) {
       fscanf(fctl, "%d", &sptree.nseqsp[i]);
       maxns += sptree.nseqsp[i];
       if(sptree.nseqsp[i]<1)  /* sptree.migration is not read yet. */
          /*puts*/error2("Shall we have at least 1 seq from each species?");
    }
    if(maxns>NS) error2("raise NS in source file");
    if(maxns<2) error2("<2 seqs?  Too simple to do anything about.");
//    printf("%d species: ", sptree.nspecies);
//    for(i=0; i<sptree.nspecies; i++)
//       printf(" %s (%d)", com.spname[i], sptree.nseqsp[i]);
//    putchar('\n');

//    data.maxns = maxns;
    fgets(line, lline, fctl);
    if(sptree.nspecies==1) {
//       sptree.root = sptree.nodes[0].nson = 0;  sptree.nnode = 1; sptree.nbranch = 0;
//       sptree.nodes[0].age = 0;
//       sptree.npop = 1;
//       sptree.pops[0] = 0;
//       sptree.pptable[0][0] = (char)1;
    }
    else {
       fgets(line, lline, fctl);
       treeString = QString::fromLocal8Bit(line);
//       ReadTreeN (fctl, &i, &j, 0, 1);
//       OutTreeN(F0,1,com.simulation);
//       FPN(F0);

//       /* copy into sptree */
//       sptree.nnode = tree.nnode;
//       sptree.nbranch = tree.nbranch;
//       sptree.root=tree.root;
//       for(i=0; i<sptree.nnode; i++) {
//          sptree.nodes[i].father = nodes[i].father;
//          sptree.nodes[i].nson = nodes[i].nson;
//          for(j=0;j<sptree.nodes[i].nson;j++)
//             sptree.nodes[i].sons[j] = nodes[i].sons[j];
//          sptree.nodes[i].age = nodes[i].age = nodes[i].branch;
//          sptree.nodes[i].theta = nodes[i].label;
//       }
//       for(i=0,sptree.npop=0; i<sptree.nspecies; i++)
//          if(sptree.nseqsp[i]>1)
//             sptree.pops[sptree.npop++] = i;
//       for(; i<2*sptree.nspecies-1; i++)
//          sptree.nodes[i].name[0] = '\0';

//       DownSptreeSetPops(sptree.root, 1, 1);
//       SetupPopPopTable(1);
//    }

//    if(com.simulation) {
//       for(i=0,com.ns=0; i<sptree.nspecies; i++)
//          com.ns += sptree.nseqsp[i];
//       if(com.ns>NS)
//          error2("raise NS in bpp.c & recompile");
//    }
//    else {
//       printf("\n%d theta parameters (populations) in the order:\n ", sptree.npop);
//       for(i=0; i<sptree.npop; i++)
//          if(i<sptree.nspecies || sptree.nspecies<=10)
//             printf(" theta_%d%s", sptree.pops[i]+1, sptree.nodes[sptree.pops[i]].name);
//          else
//             printf(" theta_%d", sptree.pops[i]+1);
//       if(sptree.nspecies>1) {
//          printf("\n%d species divergence times in the order:\n ", sptree.nspecies-1);
//          for(i=sptree.nspecies; i<sptree.nspecies*2-1; i++)
//             if(sptree.nspecies<=10) printf(" tau_%d%s", i+1, sptree.nodes[i].name);
//             else                    printf(" tau_%d", i+1);
//       }
//       printf("\n");
    }
    return(0);
}
