#ifndef CSTYLEBPPCTLLOADER_H
#define CSTYLEBPPCTLLOADER_H

//----- C Style code for loading bpp ctl file in bpp:  paml.h tools.c bpp.c -----

#include <QString>

#define MAXNFIELDS 10000

#define NSPECIES      47         /* max # of species */
#define NS            200        /* max # of sequences */
#define NBRANCH       NS*2-2     /* max # of branches.       Don't change this */
#define MAXNSONS      2          /* max # of sons per node.  Don't change this */
#define NGENE         20000      /* max # of loci */
#define LSPNAME       4000       /* max # characters in a sequence name */

class BPPEngine;

class CStyleBPPCtlLoader
{
public:
    CStyleBPPCtlLoader(BPPEngine* engine) : engine(engine) {}
    int GetOptions(char *ctlf);
    void error2(const char * message);
    int splitline(char line[], int fields[]);
    int ReadSpeciesTree (FILE* fctl, char *currline);

public:
    struct CommonInfo {
       unsigned char *z[2*NS-1];
       char *spname[NS], seqf[512], Imapf[512],outf[512], mcmcf[512], locusratef[512], heredityf[512];
       char cleandata, oldconP[NS*2-1];
       int ns, ls, lgene[1], model, clock, simulation, SeqFileFormat;
       int seqtype, ngene, *pose, npatt, np, readpattern;
       int ntime, ncatG, ncode, fix_rgene, posG[2];
       double alpha, pi[4], piG[1][4], *rates, rgene[1];
       double *conP;  /* not used */
       int curconP;
       size_t sconP;
       double *conPin[2], *fpatt, space[1000000];  /* change space to dynamic memory? */
       double a_locusrate;  /* a_locusrate is duplicated in com for simulation & in data. */
    }  com;

    struct SPECIESTREE {
       int nbranch, nnode, root, nspecies, nseqsp[NSPECIES];
       int npop, pops[NSPECIES*2-1];
       signed char pptable[NSPECIES*2-1][NSPECIES*2-1];
       int speciesdelimitation, speciestree, SpeciesModelPrior;
       int migration;
       double M[NSPECIES*2-1][NSPECIES*2-1];
       double roottau;  /* dist is seq distance */
       int nModels, iModel;  /* for species delimitation */
       char *DelimitationModels, *ModelAllowed;  /* for species delimitation */
       double *pmodel;
       struct TREESPN {
          char name[LSPNAME*2];
          int father, nson, sons[2];
          double age, theta, prob1;       /* age is tau.  prob1 is prob that the node is split. */
       } nodes[2*NSPECIES-1];
    }  sptree, sptreet;

    enum {UniformLH, UniformRootedTree, UserNodeProb} SPECIESMODELPRIORS;

    struct DATA { /* locus-specific data and gene tree information, used in lnpGB */
       int maxns, ngene, lgene[NGENE];
       int ns[NGENE], nseqsp[NGENE][NSPECIES], ls[NGENE], npatt[NGENE];
       int root[NGENE+1], conP_offset[NGENE];
       unsigned char *z[NGENE][NS];
       char cleandata[NGENE], *Imap, *Inames[NS];
       double *fpatt[NGENE], *lnpGBi, *lnpDi, lnpSpeciesModel, *locusrate, *heredity;
       double theta_prior[2], tau_prior[3];
       double a_locusrate, a_heredity, b_heredity;
       double a_seqerr[NSPECIES][4*4], e_seqerr[NSPECIES][4*4];
       int    est_locusrate, est_heredity, nseqerr, iseqerr[NSPECIES];
    }  data;

    struct MCMCPARAMETERS {
       int resetFinetune, burnin, nsample, sampfreq, usedata, saveconP;
       int print, printGenetree, printlocusrate, printheredity, moveinnode, RJalgorithm;
       double finetune[7], RJfinetune[2];
    }  mcmc;                /* control parameters */

    int seed;
    int noisy;

    QString seqfile;
    QString Imapfile;
    QString outfile;
    QString mcmcfile;
    QString treeString;
    QString locusratefile;
    QString heredityfile;

private:
    BPPEngine* engine;
};

#endif // CSTYLEBPPCTLLOADER_H
