//------------------------------------------------
// Copyright (C) 2012 Bo Xu <xuxbob@gmail.com>
//------------------------------------------------

#include "global.h"
#include <QFile>
#include <QTextStream>

QString getVersionByPath(const QString& aPath)
{
    QString version("");
    QFile codeFile( aPath + "/bpp.c" );

    if( codeFile.open(QIODevice::ReadOnly) ) {
        QTextStream codeIn(&codeFile);

        while(!codeIn.atEnd()) {
            QString line = codeIn.readLine();
            if( line.trimmed().startsWith("char VerStr") ) {
                version = line.trimmed().section(' ', 2, 2).section(',', 0, 0);
                break;
            }
        }
    }

    return version;
}

int compareVersion(const QString& version1, const QString& version2)
{
    unsigned int major1, major2, minor1, minor2;
    QString revision1, revision2;
    char c;

    QTextStream(version1.toAscii()) >> major1 >> c >> minor1 >> revision1;
    QTextStream(version2.toAscii()) >> major2 >> c >> minor2 >> revision2;

    int ret;

    if(major1 > major2)
        ret = 1;
    else if(major1 < major2)
        ret = -1;
    else {
        if(minor1 > minor2)
            ret = 1;
        else if(minor1 < minor2)
            ret = -1;
        else
            ret = QString::compare(revision1, revision2, Qt::CaseInsensitive);
    }

    return ret;
}
