//------------------------------------------------
// Copyright (C) 2012 Bo Xu <xuxbob@gmail.com>
//------------------------------------------------

#include "mainwindow.h"
#include "bppengine.h"
#include "configurationdialog.h"
#include "bpphistorydialog.h"
#include "outputwindow.h"
#include "spinboxdelegate.h"
#include <QCompleter>
#include <QFileDialog>
#include <QFileSystemModel>
#include <QCloseEvent>
#include <QClipboard>
#include <QMessageBox>
#include <QPrinter>
#include <QFontDialog>
#include <QPrintPreviewDialog>
#include <QPrintDialog>
#include <QUrl>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
#if defined(Q_OS_MAC)
    defaultOutputFont("Courier New", 12),
    defaultPrintFont("Courier New", 14),
#elif defined(Q_OS_LINUX)
    defaultOutputFont("Liberation Mono", 10),
    defaultPrintFont("Sans Serif", 12),
#else
    defaultOutputFont("Courier New", 8),
    defaultPrintFont("Courier New", 9),
#endif
    bppEngine(NULL),
    bppOutput(NULL),
    currentDocument(NULL),
    focusEdit(NULL),
    actionGroup(NULL)
{
    setupUi(this);

    setAcceptDrops(true);
    setWindowTitle(QApplication::applicationName());

    //Menu
    QAction* actionBefore = menuRecent_Files->actions().first();
    for(int i=MaxRecentFiles-1; i>=0; i--) {
        recentFileActions[i] = new QAction(this);
        recentFileActions[i]->setData(i);
        recentFileActions[i]->setVisible(false);
        recentFileActions[i]->setStatusTip(tr("Open this file"));
        if(i < MaxRecentFiles-1)
            actionBefore = recentFileActions[i+1];
        menuRecent_Files->insertAction(actionBefore, recentFileActions[i]);
        QObject::connect(recentFileActions[i], SIGNAL(triggered()), this, SLOT(onActionOpenRecentFile()));
    }

    actionNew->setShortcut(QKeySequence::New);
    actionOpen->setShortcut(QKeySequence::Open);
    actionClose->setShortcut(QKeySequence::Close);
    actionSave->setShortcut(QKeySequence::Save);
    actionSave_As->setShortcut(QKeySequence::SaveAs);
    actionPrint->setShortcut(QKeySequence::Print);

    actionUndo->setShortcut(QKeySequence::Undo);
    actionRedo->setShortcut(QKeySequence::Redo);
    actionCut->setShortcut(QKeySequence::Cut);
    actionCopy->setShortcut(QKeySequence::Copy);
    actionPaste->setShortcut(QKeySequence::Paste);
    actionDelete->setShortcut(QKeySequence::Delete);
    actionSelect_All->setShortcut(QKeySequence::SelectAll);

    actionText_View->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_T));

    actionBPP_User_Guide->setShortcut(QKeySequence::HelpContents);

    installEventFilter(this);

#ifdef Q_OS_MAC
    QIcon icon;
    icon.addFile(QString::fromUtf8(":/empty.png"), QSize(), QIcon::Normal, QIcon::Off);
    actionClose->setIcon(icon);
    actionSave_As->setIcon(icon);
    actionSet_Printer_Font->setIcon(icon);
    actionPrint_Preview->setIcon(icon);
    menuRecent_Files->setIcon(icon);
    actionSelect_All->setIcon(icon);

    actionConfiguration->setText(QApplication::translate("MainWindow", "&Preferences...", 0, QApplication::UnicodeUTF8));
#endif

    actionGroup = new QActionGroup(this);
    actionGroup->addAction(actionStart_Page);
    actionGroup->addAction(actionBPP);

    //Stacked widget
#ifdef Q_OS_MAC
    QFont font;
    font.setPointSize(13);
    stackedWidget->setFont(font);
#endif
    QFont mainWidgetFont = stackedWidget->font();

    //Start page
    actionStart_Page->trigger();

#ifdef Q_OS_MAC
    label_72->setText(QApplication::translate("MainWindow", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:'Lucida Grande'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:9pt;\"></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:9pt;\"></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:9pt;\"></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:'Chalkduster'; font-size:28pt; font-weight:600;\">   </span><img src=\":/bppxlogo.png\" style=\"vertical-align: middle;\" /><span style=\" font-family:'Chalkduster'; font-size:28pt; font-weight:600;\">    Bayesian Phylogeny and Phylogeography          </span></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:10pt; font-weight:600;\"></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:10pt; font-weight:600;\"></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://abacus.gene.ucl.ac.uk/\"><span style=\" font-family:'Nanum Pen Script'; font-size:36pt; font-weight:600; text-decoration: none; color:#000000;\">Ziheng Yang</span></a></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:9pt;\"></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:9pt;\"></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://abacus.gene.ucl.ac.uk/software.html\"><span style=\" font-family:'Herculanum'; font-size:20pt; text-decoration: none; color:#000000;\">BP&P Software Website</span></a></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:20pt; font-weight:600;\"></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'SimSun'; font-size:9pt;\"></p></body></html>", 0, QApplication::UnicodeUTF8));

    QFont font1;
    font1.setFamily(QString::fromUtf8("Comic Sans MS"));
    font1.setPointSize(14);
    font1.setBold(true);
    font1.setItalic(false);
    font1.setWeight(75);
    commandLinkButton->setFont(font1);

    QIcon icon1;
    icon1.addFile(QString::fromUtf8(":/arrow.png"), QSize(), QIcon::Normal, QIcon::Off);
    commandLinkButton->setIcon(icon1);
#endif

    //bpp
    tableWidget->setItemDelegateForColumn(1, new SpinBoxDelegate());

    QCompleter* completer_20 = new QCompleter(this);
    completer_20->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
#ifdef Q_OS_WIN32
    completer_20->setCaseSensitivity(Qt::CaseInsensitive);
#endif
    QFileSystemModel* fileSystemModel_20 = new QFileSystemModel(completer_20);
    fileSystemModel_20->setRootPath("");
    completer_20->setModel(fileSystemModel_20);
    comboBox_20->setCompleter(completer_20);

    QCompleter* completer_21 = new QCompleter(this);
    completer_21->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
#ifdef Q_OS_WIN32
    completer_21->setCaseSensitivity(Qt::CaseInsensitive);
#endif
    QFileSystemModel* fileSystemModel_21 = new QFileSystemModel(completer_21);
    fileSystemModel_21->setRootPath("");
    completer_21->setModel(fileSystemModel_21);
    comboBox_21->setCompleter(completer_21);

    QCompleter* completer_22 = new QCompleter(this);
    completer_22->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
#ifdef Q_OS_WIN32
    completer_22->setCaseSensitivity(Qt::CaseInsensitive);
#endif
    QFileSystemModel* fileSystemModel_22 = new QFileSystemModel(completer_22);
    fileSystemModel_22->setRootPath("");
    QDir::Filters filter_22 = fileSystemModel_22->filter();
    fileSystemModel_22->setFilter( filter_22 & ~QDir::Files );
    completer_22->setModel(fileSystemModel_22);
    comboBox_22->setCompleter(completer_22);

    QCompleter* completer_23 = new QCompleter(this);
    completer_23->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
#ifdef Q_OS_WIN32
    completer_23->setCaseSensitivity(Qt::CaseInsensitive);
#endif
    QFileSystemModel* fileSystemModel_23 = new QFileSystemModel(completer_23);
    fileSystemModel_23->setRootPath("");
    completer_23->setModel(fileSystemModel_23);
    comboBox_23->setCompleter(completer_23);

    QCompleter* completer_24 = new QCompleter(this);
    completer_24->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
#ifdef Q_OS_WIN32
    completer_24->setCaseSensitivity(Qt::CaseInsensitive);
#endif
    QFileSystemModel* fileSystemModel_24 = new QFileSystemModel(completer_24);
    fileSystemModel_24->setRootPath("");
    completer_24->setModel(fileSystemModel_24);
    comboBox_24->setCompleter(completer_24);

    comboBox_20->setFont(mainWidgetFont);
    comboBox_21->setFont(mainWidgetFont);
    comboBox_22->setFont(mainWidgetFont);
    comboBox_23->setFont(mainWidgetFont);
    comboBox_24->setFont(mainWidgetFont);

    comboBox_20->lineEdit()->setDragEnabled(true);
    comboBox_21->lineEdit()->setDragEnabled(true);
    comboBox_22->lineEdit()->setDragEnabled(true);
    comboBox_23->lineEdit()->setDragEnabled(true);
    comboBox_24->lineEdit()->setDragEnabled(true);

    lineEdit_6->setFont(mainWidgetFont);

    listWidget->setFont(mainWidgetFont);


    loadSettings();

    updateEnginesPath();
    updateOutputFont();
    updateRecentFileActions();
}

MainWindow::~MainWindow()
{
    focusEdit = NULL;

    delete bppOutput;
    bppOutput = NULL;

    delete bppEngine;
    bppEngine = NULL;

    currentDocument = NULL;
}

bool MainWindow::eventFilter(QObject* watched, QEvent* event)
{
    if( event->type() == QEvent::ShortcutOverride ) {
        QKeyEvent* key = static_cast<QKeyEvent*>(event);
        if(key) {
            if( key->matches(QKeySequence::New) )
                actionNew->setEnabled( !actionText_View->isChecked() );
            else if( key->matches(QKeySequence::Open) )
                actionOpen->setEnabled( !actionText_View->isChecked() );
            else if( key->matches(QKeySequence::Close) )
                actionClose->setEnabled( currentDocument && !currentDocument->title().isEmpty() && !actionText_View->isChecked() );
            else if( key->matches(QKeySequence::Save) )
                actionSave->setEnabled( currentDocument && currentDocument->isModified() );
            else if( key->matches(QKeySequence::SaveAs) )
                actionSave_As->setEnabled( currentDocument );
            else if( key->matches(QKeySequence::Print) )
                actionPrint->setEnabled( currentDocument && !currentDocument->title().isEmpty() );
            else if( key->modifiers() == Qt::ControlModifier && key->key() == Qt::Key_T )
                actionText_View->setEnabled( currentDocument && !currentDocument->title().isEmpty() );
        }
    }

    return QMainWindow::eventFilter(watched, event);
}

void MainWindow::closeEvent(QCloseEvent* event)
{
    if(bppEngine) {
        if( !saveModifiedFile(bppEngine) ) {
            event->ignore();
            return;
        }
    }

    delete bppOutput;
    bppOutput = NULL;

    saveSettings();
    event->accept();
}

void MainWindow::dragEnterEvent(QDragEnterEvent* event)
{
    if( event->mimeData()->hasFormat("text/uri-list") )
        event->acceptProposedAction();
}

void MainWindow::dropEvent(QDropEvent* event)
{
    QList<QUrl> urls = event->mimeData()->urls();
    if(urls.isEmpty()) return;

    QString fileName = urls.first().toLocalFile();
    if(fileName.isEmpty()) return;

    loadDocumentFile(fileName);
    event->acceptProposedAction();
}

void MainWindow::loadSettings()
{
    QString settingsFileName = QApplication::applicationDirPath();
    settingsFileName += "/bppx.conf";
    QFile settingsFile(settingsFileName);

    if( settingsFile.open(QIODevice::ReadOnly) ) {
        QDataStream settingsIn(&settingsFile);
        settingsIn.setVersion(QDataStream::Qt_4_0);

        const quint32 magicNumber = 0x30A6FF4E;
        quint32 storedMagicNumber;
        settingsIn >> storedMagicNumber;

        if(storedMagicNumber == magicNumber) {
            QByteArray geometry;
            bool statusBarVisible;

            settingsIn >> bppPath
                       >> outputFont
                       >> printFont
                       >> recentFiles
                       >> lastFileOpenPath
                       >> geometry
                       >> statusBarVisible;

            restoreGeometry(geometry);
            QMainWindow::statusBar()->setVisible(statusBarVisible);

            if( QDir::isAbsolutePath(bppPath) ) {
                absoluteBPPPath = bppPath;
            }
            else {
                QString appPath = QApplication::applicationDirPath();
#ifdef Q_OS_MAC
                QDir dir(appPath);
                dir.cd("../../..");
                appPath = dir.absolutePath();
#endif
                absoluteBPPPath = QDir::cleanPath( appPath + "/" + bppPath );
            }

            return;
        }
    }

//DefaultSettings
    bppPath = QApplication::applicationDirPath();
#ifdef Q_OS_MAC
    QDir dir(bppPath);
    dir.cd("../../..");
    bppPath = dir.absolutePath();
#endif
    outputFont = defaultOutputFont;
    printFont = defaultPrintFont;
    lastFileOpenPath = "";
    absoluteBPPPath = bppPath;
}

void MainWindow::saveSettings()
{
    QString settingsFileName = QApplication::applicationDirPath();
    settingsFileName += "/bppx.conf";
    QFile settingsFile(settingsFileName);

    if( settingsFile.open(QIODevice::WriteOnly) ) {
        QDataStream settingsOut(&settingsFile);
        settingsOut.setVersion(QDataStream::Qt_4_0);

        const quint32 magicNumber = 0x30A6FF4E;
        settingsOut << magicNumber
                    << bppPath
                    << outputFont
                    << printFont
                    << recentFiles
                    << lastFileOpenPath
                    << saveGeometry()
                    << QMainWindow::statusBar()->isVisible();
    }
}

void MainWindow::loadDocumentFile(const QString& fileName)
{
    actionBPP->trigger();

    //if( !fileName.isEmpty() && fileName == currentDocument->fileName() ) return;

    if( !saveModifiedFile(currentDocument) ) return;

    if(fileName.isEmpty()) {
        //NemDocument
        if( currentDocument->onNewDocument() ) {
            updateUi();
            updateUiTitle();
        }
    }
    else {
        //OpenDocument
        if( currentDocument->onOpenDocument(fileName) ) {
            updateUi();
            updateUiTitle();
            addToRecentFiles(fileName);
        }
    }
}

bool MainWindow::saveDocumentFile(const QString& fileName)
{
    bool hadDoc = !currentDocument->title().isEmpty();

    bool ret = currentDocument->onSaveDocument(fileName);

    if(ret) {
        if(!hadDoc) updateUi();
        updateUiTitle();

        addToRecentFiles(fileName);
    }

    return ret;
}

void MainWindow::addToRecentFiles(const QString& fileName)
{
    if(fileName.isEmpty()) return;

    recentFiles.removeAll(fileName);
    recentFiles.prepend(fileName);
    updateRecentFileActions();
}

void MainWindow::removeFromRecentFiles(const QString& fileName)
{
    if(fileName.isEmpty()) return;

    if(recentFiles.removeAll(fileName) > 0)
        updateRecentFileActions();
}

bool MainWindow::saveModifiedFile(Document* document, bool saveInBackground)
{
    bool retVal = true;

    if( document && document->isModified() && !document->title().isEmpty() ) {
        Document* oldDocument = currentDocument;

        if(document != currentDocument && !saveInBackground)
            actionBPP->trigger();

        QString name = document->fileName();
        if(name.isEmpty())
            name = document->title();

        QMessageBox msgBox;
        msgBox.setIcon(QMessageBox::Question);
        msgBox.setWindowTitle(tr("Save Changes"));
        msgBox.setText(tr("Save changes to:"));
        msgBox.setInformativeText(QDir::toNativeSeparators(name));
        msgBox.addButton(QMessageBox::Yes);
        msgBox.addButton(QMessageBox::No);
        msgBox.addButton(QMessageBox::Cancel);
        msgBox.setDefaultButton(QMessageBox::Yes);
        int ret = msgBox.exec();

        if(ret == QMessageBox::Yes) {
            bool ret2;

            if(document != currentDocument && saveInBackground) {
                currentDocument = document;
                ret2 = on_actionSave_triggered();
                currentDocument = oldDocument;

                if(ret2) updateUiTitle();
            }
            else
                ret2 = on_actionSave_triggered();

            if(!ret2) retVal = false;
        }
        else if(ret == QMessageBox::Cancel) {
            retVal = false;
        }

        if(oldDocument != currentDocument && !saveInBackground && !retVal)
            actionStart_Page->trigger();
    }

    return retVal;
}

void MainWindow::setModifiedFlag()
{
    if( currentDocument && !currentDocument->isModified() && !currentDocument->title().isEmpty() ) {
        currentDocument->setModified(true);
        setWindowModified(true);
    }
}

void MainWindow::updateEnginesPath()
{
    if(bppEngine) {
        bppEngine->setPath(absoluteBPPPath);
        if( bppEngine->state() == Engine::NotRunning )
            pushButton->setEnabled( bppEngine->isStatusOk() );
    }
}

void MainWindow::updateOutputFont()
{
    //QTextEdit font
    if(bppOutput)
        bppOutput->textEdit->setCurrentFont(outputFont);
}

void MainWindow::updateRecentFileActions()
{
    int i=0;
    int size = recentFiles.size();

    for(; i<MaxRecentFiles && i<size; i++) {
        QString text = tr("&%1 %2").arg(i+1).arg(QDir::toNativeSeparators(recentFiles[i]));
        recentFileActions[i]->setText(text);
        recentFileActions[i]->setVisible(true);
    }

    for(; i<MaxRecentFiles; i++) {
        recentFileActions[i]->setVisible(false);
    }
}

void MainWindow::updateUiTitle()
{
    QString appName = QApplication::applicationName();

    QString fileTitle;
    if(currentDocument)
        fileTitle = currentDocument->title();

    if(fileTitle.isEmpty()) {
        setWindowTitle(appName);
        setWindowModified(false);
    }
    else if(!actionText_View->isChecked()) {
        setWindowTitle( tr("%1  -  [%2[*]]").arg(appName).arg(fileTitle) );
        setWindowModified( currentDocument->isModified() );
    }
    else {
        setWindowTitle( tr("%1  -  [%2[*]] - Text View (ReadOnly)").arg(appName).arg(fileTitle) );
        setWindowModified( currentDocument->isModified() );
    }
}

void MainWindow::updateUi()
{
    if(!bppEngine) return;

    lineEdit_5->setText( QString::number(bppEngine->nEdit_seed) );
    int i = bppEngine->nComb_noisy;
    if(i > 4 && i < 9) i = 4;
    else if(i >= 9) i = 5;
    comboBox->setCurrentIndex(i);
    comboBox_20->lineEdit()->setText( bppEngine->strEdit_seqfile );
    comboBox_21->lineEdit()->setText( bppEngine->strEdit_Imapfile );
    lineEdit_3->setText( bppEngine->strEdit_outfile_Name );
    lineEdit_4->setText( bppEngine->strEdit_mcmcfile_Name );
    comboBox_22->lineEdit()->setText( bppEngine->strEdit_Output_Location );
    comboBox_2->setCurrentIndex( bppEngine->nComb_speciesdelimitation );
    comboBox_3->setCurrentIndex( bppEngine->nComb_RJalgorithm );
    doubleSpinBox->setValue( bppEngine->nEdit_RJfinetune[0] );
    doubleSpinBox_2->setValue( bppEngine->nEdit_RJfinetune[1] );
    comboBox_8->setCurrentIndex( bppEngine->nComb_speciestree );
    comboBox_4->setCurrentIndex( bppEngine->nComb_speciesmodelprior );
    lineEdit->setText( QString::number(bppEngine->nEdit_nspecies) );

    int rowCount = tableWidget->rowCount();

    if( rowCount > bppEngine->nEdit_nspecies ) {
        for(int i=rowCount-1; i>=bppEngine->nEdit_nspecies; i--) {
            tableWidget->removeRow(i);
        }
        rowCount = bppEngine->nEdit_nspecies;
    }

    tableWidget->setRowCount( bppEngine->nEdit_nspecies );

    const bool sortingEnabled = tableWidget->isSortingEnabled();
    tableWidget->setSortingEnabled(false);
    for(i=0; i<rowCount; i++) {
        tableWidget->item(i, 0)->setText( bppEngine->cList_species[i].name );
        tableWidget->item(i, 1)->setText( QString::number(bppEngine->cList_species[i].nseqsp) );
    }
    for(i=rowCount; i < bppEngine->nEdit_nspecies; i++) {
        tableWidget->setItem(i, 0, new QTableWidgetItem( bppEngine->cList_species[i].name ));
        tableWidget->setItem(i, 1, new QTableWidgetItem( QString::number(bppEngine->cList_species[i].nseqsp) ));
    }
    tableWidget->setSortingEnabled(sortingEnabled);

    lineEdit_6->setText( bppEngine->strEdit_tree );
    comboBox_5->setCurrentIndex( bppEngine->nComb_usedata );
    spinBox_2->setValue( bppEngine->nEdit_nloci );
    checkBox_2->setChecked( bppEngine->bButt_cleandata );
    doubleSpinBox_6->setValue( bppEngine->nEdit_thetaprior[0] );
    doubleSpinBox_7->setValue( bppEngine->nEdit_thetaprior[1] );
    doubleSpinBox_8->setValue( bppEngine->nEdit_tauprior[0] );
    doubleSpinBox_9->setValue( bppEngine->nEdit_tauprior[1] );
    doubleSpinBox_10->setValue( bppEngine->nEdit_tauprior[2] );
    comboBox_6->setCurrentIndex( bppEngine->nComb_locusrate );
    doubleSpinBox_11->setValue( bppEngine->nEdit_a_locusrate );
    comboBox_23->lineEdit()->setText( bppEngine->strEdit_locusratefile );
    comboBox_7->setCurrentIndex( bppEngine->nComb_heredity );
    doubleSpinBox_12->setValue( bppEngine->nEdit_a_heredity );
    doubleSpinBox_13->setValue( bppEngine->nEdit_b_heredity );
    comboBox_24->lineEdit()->setText( bppEngine->strEdit_heredityfile );
    checkBox->setChecked( bppEngine->bButt_autoAdjustFinetune );
    doubleSpinBox_14->setValue( bppEngine->nEdit_finetune[0] );
    doubleSpinBox_15->setValue( bppEngine->nEdit_finetune[1] );
    doubleSpinBox_16->setValue( bppEngine->nEdit_finetune[2] );
    doubleSpinBox_17->setValue( bppEngine->nEdit_finetune[3] );
    doubleSpinBox_18->setValue( bppEngine->nEdit_finetune[4] );
    doubleSpinBox_19->setValue( bppEngine->nEdit_finetune[5] );
    doubleSpinBox_20->setValue( bppEngine->nEdit_finetune[6] );
    checkBox_3->setChecked( bppEngine->bButt_print );
    checkBox_4->setChecked( bppEngine->bButt_printlocusrate );
    checkBox_5->setChecked( bppEngine->bButt_printheredity );
    checkBox_6->setChecked( bppEngine->bButt_printGenetree );
    spinBox_3->setValue( bppEngine->nEdit_burnin );
    spinBox_4->setValue( bppEngine->nEdit_sampfreq );
    spinBox_5->setValue( bppEngine->nEdit_nsample );
}

void MainWindow::updateEngine()
{
    if(!bppEngine) return;

    bppEngine->nEdit_seed = lineEdit_5->text().toInt();
    int i = comboBox->currentIndex();
    if(i == 5) i = 9;
    bppEngine->nComb_noisy = i;
    bppEngine->strEdit_seqfile = comboBox_20->lineEdit()->text();
    bppEngine->strEdit_Imapfile = comboBox_21->lineEdit()->text();
    bppEngine->strEdit_outfile_Name = lineEdit_3->text();
    bppEngine->strEdit_mcmcfile_Name = lineEdit_4->text();
    bppEngine->strEdit_Output_Location = comboBox_22->lineEdit()->text();
    bppEngine->nComb_speciesdelimitation = comboBox_2->currentIndex();
    bppEngine->nComb_RJalgorithm = comboBox_3->currentIndex();
    bppEngine->nEdit_RJfinetune[0] = doubleSpinBox->value();
    bppEngine->nEdit_RJfinetune[1] = doubleSpinBox_2->value();
    bppEngine->nComb_speciestree = comboBox_8->currentIndex();
    bppEngine->nComb_speciesmodelprior = comboBox_4->currentIndex();
    bppEngine->nEdit_nspecies = lineEdit->text().toInt();

    bppEngine->cList_species.clear();
    int rowCount = tableWidget->rowCount();
    for(int i=0; i < rowCount; i++) {
        bppEngine->cList_species[i].name = tableWidget->item(i, 0)->text();
        bppEngine->cList_species[i].nseqsp = tableWidget->item(i, 1)->text().toInt();
    }

    bppEngine->strEdit_tree = lineEdit_6->text();
    bppEngine->nComb_usedata = comboBox_5->currentIndex();
    bppEngine->nEdit_nloci = spinBox_2->value();
    bppEngine->bButt_cleandata = checkBox_2->isChecked();
    bppEngine->nEdit_thetaprior[0] = doubleSpinBox_6->value();
    bppEngine->nEdit_thetaprior[1] = doubleSpinBox_7->value();
    bppEngine->nEdit_tauprior[0] = doubleSpinBox_8->value();
    bppEngine->nEdit_tauprior[1] = doubleSpinBox_9->value();
    bppEngine->nEdit_tauprior[2] = doubleSpinBox_10->value();
    bppEngine->nComb_locusrate = comboBox_6->currentIndex();
    bppEngine->nEdit_a_locusrate = doubleSpinBox_11->value();
    bppEngine->strEdit_locusratefile = comboBox_23->lineEdit()->text();
    bppEngine->nComb_heredity = comboBox_7->currentIndex();
    bppEngine->nEdit_a_heredity = doubleSpinBox_12->value();
    bppEngine->nEdit_b_heredity = doubleSpinBox_13->value();
    bppEngine->strEdit_heredityfile = comboBox_24->lineEdit()->text();
    bppEngine->bButt_autoAdjustFinetune = checkBox->isChecked();
    bppEngine->nEdit_finetune[0] = doubleSpinBox_14->value();
    bppEngine->nEdit_finetune[1] = doubleSpinBox_15->value();
    bppEngine->nEdit_finetune[2] = doubleSpinBox_16->value();
    bppEngine->nEdit_finetune[3] = doubleSpinBox_17->value();
    bppEngine->nEdit_finetune[4] = doubleSpinBox_18->value();
    bppEngine->nEdit_finetune[5] = doubleSpinBox_19->value();
    bppEngine->nEdit_finetune[6] = doubleSpinBox_20->value();
    bppEngine->bButt_print = checkBox_3->isChecked();
    bppEngine->bButt_printlocusrate = checkBox_4->isChecked();
    bppEngine->bButt_printheredity = checkBox_5->isChecked();
    bppEngine->bButt_printGenetree = checkBox_6->isChecked();
    bppEngine->nEdit_burnin = spinBox_3->value();
    bppEngine->nEdit_sampfreq = spinBox_4->value();
    bppEngine->nEdit_nsample = spinBox_5->value();
}

void MainWindow::showWarningMessageWhenPathFailed(const QString &aPath, bool isExecutable)
{
    QMessageBox msgBox;
    msgBox.setIcon(QMessageBox::Warning);
    msgBox.setWindowTitle(tr("Warning"));
    QString msg = (isExecutable)? tr("<p>Failed&nbsp;to&nbsp;find&nbsp;executable&nbsp;file&nbsp;\"%1\"&nbsp;in&nbsp;the&nbsp;folder:<br>%2</p>") : tr("<p>Failed&nbsp;to&nbsp;find&nbsp;\"%1\"&nbsp;in&nbsp;the&nbsp;folder:<br>%2</p>");
    QFileInfo fileInfo(aPath);
    msgBox.setText(msg.arg(fileInfo.fileName()).arg(QDir::toNativeSeparators(fileInfo.path())));
#ifdef Q_OS_MAC
    msgBox.setInformativeText(tr("<p>Please&nbsp;set&nbsp;correct&nbsp;bpp&nbsp;path&nbsp;in&nbsp;<i>%1&nbsp;&rarr;&nbsp;Preferences...</i></p>").arg(QApplication::applicationName()));
#else
    msgBox.setInformativeText(tr("<p>Please&nbsp;set&nbsp;correct&nbsp;bpp&nbsp;path&nbsp;in&nbsp;<i>Tools&nbsp;&rarr;&nbsp;Configuration...</i></p>"));
#endif
    msgBox.exec();
}

void MainWindow::getSelectedRows(QList<QTableWidgetSelectionRange>& list, QList<int>& selectedRowList, bool& hasOtherRange) const
{
    int rowCount = tableWidget->rowCount();
    int columnCount = tableWidget->columnCount();
    int* array = new int[rowCount];
    for(int i=0; i<rowCount; i++)
        array[i] = 0;

    QList<QTableWidgetSelectionRange>::const_iterator it = list.begin();
    for(; it != list.end(); ++it) {
        for(int i = it->topRow(); i <= it->bottomRow(); i++)
            array[i] += it->columnCount();
    }

    hasOtherRange = false;
    for(int i=0; i<rowCount; i++) {
        if( !hasOtherRange && array[i] > 0 && array[i] < columnCount )
            hasOtherRange = true;
        else if( array[i] == columnCount )
            selectedRowList.append(i);
    }

    delete[]array;
}

bool MainWindow::checkTreeFormat(const QString& tree, int n) const {
    int k = 0;
    int m = 1;
    QString::const_iterator it = tree.begin();

    for(; it!=tree.end() && k>=0; ++it) {
        if(*it == '(') k++;
        else if(*it == ')') k--;
        else if(*it == ',') m++;
    }

    return ( (tree.isEmpty() && n == 0) || (k == 0 && m == n) );
}

void MainWindow::onBPPEngineStateChanged(Engine::EngineState newState)
{
    switch(newState)
    {
    case Engine::NotRunning:
        {
            pushButton->setEnabled( bppEngine->isStatusOk() );
            pushButton_2->setEnabled(false);
        }
        break;
    case Engine::Running:
        {
        pushButton_2->setEnabled(true);
        }
        break;
    default:;
    }
}

void MainWindow::print(QPrinter* printer)
{
    currentDocument->refreshFileContent();

#ifndef QT_NO_PRINTER
    QTextEdit* textEdit = new QTextEdit;

    textEdit->setCurrentFont(printFont);
    textEdit->setText( currentDocument->fileContent() );
    textEdit->print(printer);

    delete textEdit;
#endif
}

void MainWindow::on_stackedWidget_currentChanged(int index)
{
    actionText_View->setChecked(false);

    switch(index)
    {
    case BPP:
        {
        if(!bppEngine) {
            bppEngine = new BPPEngine(absoluteBPPPath, NULL, 5*1000, 12*3600*1000, this);
            QObject::connect(bppEngine, SIGNAL(stateChanged(Engine::EngineState)), this, SLOT(onBPPEngineStateChanged(Engine::EngineState)));

            updateUi();

            if( !bppEngine->isStatusOk() ) {
                if( bppEngine->state() == Engine::NotRunning )
                    pushButton->setEnabled(false);

                showWarningMessageWhenPathFailed( bppEngine->path(), true );
            }
        }

        currentDocument = bppEngine;
        }
        break;

    default:
        currentDocument = NULL;
    }

    updateUiTitle();
}


//----- Menu -----

void MainWindow::on_menuFile_aboutToShow()
{
    actionNew->setEnabled( !actionText_View->isChecked() );
    actionOpen->setEnabled( !actionText_View->isChecked() );
    actionClose->setEnabled( currentDocument && !currentDocument->title().isEmpty() && !actionText_View->isChecked() );
    actionSave->setEnabled( currentDocument && currentDocument->isModified() );
    actionSave_As->setEnabled( currentDocument );
    actionSet_Printer_Font->setEnabled( currentDocument && !currentDocument->title().isEmpty() );
    actionPrint_Preview->setEnabled( currentDocument && !currentDocument->title().isEmpty() );
    actionPrint->setEnabled( currentDocument && !currentDocument->title().isEmpty() );
    menuRecent_Files->setEnabled( !recentFiles.isEmpty() && !actionText_View->isChecked() );
}

void MainWindow::on_actionNew_triggered()
{
    loadDocumentFile(QString());
}

void MainWindow::on_actionOpen_triggered()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open %1 Control File").arg("BPP"), lastFileOpenPath);
    if(fileName.isEmpty()) return;

    lastFileOpenPath = QFileInfo(fileName).absolutePath();
    loadDocumentFile(fileName);
}

void MainWindow::on_actionClose_triggered()
{
    if( !saveModifiedFile(currentDocument) ) return;

    if( currentDocument->onCloseDocument() ) {
        updateUi();
        actionStart_Page->trigger();
    }
}

bool MainWindow::on_actionSave_triggered()
{
    if( currentDocument->fileName().isEmpty() )
        return on_actionSave_As_triggered();
    else
        return saveDocumentFile( currentDocument->fileName() );
}

bool MainWindow::on_actionSave_As_triggered()
{
    QString path = currentDocument->fileName();
    if(path.isEmpty())
        path = currentDocument->title();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Save As"), path);
    if(fileName.isEmpty()) return false;

    return saveDocumentFile(fileName);
}

void MainWindow::on_actionSet_Printer_Font_triggered()
{
    printFont = QFontDialog::getFont(0, printFont, this);
}

void MainWindow::on_actionPrint_Preview_triggered()
{
#ifndef QT_NO_PRINTER
    QPrinter printer;
    QPrintPreviewDialog printPreviewDialog(&printer, this);
    QObject::connect(&printPreviewDialog, SIGNAL(paintRequested(QPrinter*)), this, SLOT(print(QPrinter*)));

    printPreviewDialog.exec();
#endif
}

void MainWindow::on_actionPrint_triggered()
{
#ifndef QT_NO_PRINTER
    QPrinter printer;
    QPrintDialog printDialog(&printer, this);
    if( printDialog.exec() == QDialog::Accepted ) {
        print(&printer);
    }
#endif
}

void MainWindow::onActionOpenRecentFile()
{
    QAction* action = qobject_cast<QAction*>(sender());
    if(action) {
        int i = action->data().toInt();
        if( QFileInfo(recentFiles[i]).isFile() )
            loadDocumentFile(recentFiles[i]);
        else {
            QMessageBox msgBox;
            msgBox.setIcon(QMessageBox::Question);
            msgBox.setWindowTitle(tr("Failed to Open File"));
            msgBox.setText(tr("The file \"%1\" cannot be opened.").arg(QDir::toNativeSeparators(recentFiles[i])));
            msgBox.setInformativeText(tr("Do you want to remove the reference to it from <i>Recent Files</i> list?"));
            msgBox.addButton(QMessageBox::Yes);
            msgBox.addButton(QMessageBox::No);
            msgBox.setDefaultButton(QMessageBox::Yes);
            if( msgBox.exec() == QMessageBox::Yes )
                removeFromRecentFiles(recentFiles[i]);
        }
    }
}

void MainWindow::on_actionClear_History_triggered()
{
    recentFiles.clear();
    updateRecentFileActions();
}

void MainWindow::on_menuEdit_aboutToShow()
{
    QComboBox* aComboBox;
    QLineEdit* focusLineEdit;
    QTextEdit* focusTextEdit;
    bool result;

    focusLineEdit = qobject_cast<QLineEdit*>( QApplication::focusWidget() );
    if(focusLineEdit) {
        goto UpdateEditMenuforLineEdit;
    }

    aComboBox = qobject_cast<QComboBox*>( QApplication::focusWidget() );
    if( aComboBox && aComboBox->isEditable() ) {
        focusLineEdit = aComboBox->lineEdit();
        goto UpdateEditMenuforLineEdit;
    }

    focusTextEdit = qobject_cast<QTextEdit*>( QApplication::focusWidget() );
    if(focusTextEdit) {
        goto UpdateEditMenuforTextEdit;
    }

    actionUndo->setEnabled(false);
    actionRedo->setEnabled(false);
    actionCut->setEnabled(false);
    actionCopy->setEnabled(false);
    actionPaste->setEnabled(false);
    actionDelete->setEnabled(false);
    actionSelect_All->setEnabled(false);

    return;


UpdateEditMenuforLineEdit:

    result = focusLineEdit->hasSelectedText();

    actionUndo->setEnabled( focusLineEdit->isUndoAvailable() );
    actionRedo->setEnabled( focusLineEdit->isRedoAvailable() );
    actionCut->setEnabled(result);
    actionCopy->setEnabled(result);
    actionPaste->setEnabled( !QApplication::clipboard()->text().isEmpty() );
    actionDelete->setEnabled(result);
    actionSelect_All->setEnabled( !focusLineEdit->text().isEmpty() );

    focusEdit = focusLineEdit;

    return;


UpdateEditMenuforTextEdit:

    actionUndo->setEnabled(false);
    actionRedo->setEnabled(false);
    actionCut->setEnabled(false);
    actionCopy->setEnabled( !focusTextEdit->textCursor().selectedText().isEmpty() );
    actionPaste->setEnabled(false);
    actionDelete->setEnabled(false);
    actionSelect_All->setEnabled( !focusTextEdit->toPlainText().isEmpty() );

    focusEdit = focusTextEdit;

    return;
}

void MainWindow::on_actionUndo_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit)
        focusLineEdit->undo();
}

void MainWindow::on_actionRedo_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit)
        focusLineEdit->redo();
}

void MainWindow::on_actionCut_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit)
        focusLineEdit->cut();
}

void MainWindow::on_actionCopy_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit) {
        focusLineEdit->copy();
        return;
    }

    QTextEdit* focusTextEdit = qobject_cast<QTextEdit*>(focusEdit);
    if(focusTextEdit) {
        focusTextEdit->copy();
        return;
    }
}

void MainWindow::on_actionPaste_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit)
        focusLineEdit->paste();
}

void MainWindow::on_actionDelete_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit)
        focusLineEdit->del();
}

void MainWindow::on_actionSelect_All_triggered()
{
    QLineEdit* focusLineEdit = qobject_cast<QLineEdit*>(focusEdit);
    if(focusLineEdit) {
        focusLineEdit->selectAll();
        return;
    }

    QTextEdit* focusTextEdit = qobject_cast<QTextEdit*>(focusEdit);
    if(focusTextEdit) {
        focusTextEdit->selectAll();
        return;
    }
}

void MainWindow::on_menuFunction_aboutToShow()
{
    actionText_View->setEnabled( currentDocument && !currentDocument->title().isEmpty() );
    actionStatus_Bar->setChecked( QMainWindow::statusBar()->isVisible() );

    actionOutput_Window->setEnabled( bppOutput );
    actionOutput_Window->setChecked( bppOutput && bppOutput->isVisible() );
}

void MainWindow::on_actionStart_Page_triggered()
{
    stackedWidget->setCurrentIndex(StartPage);
}

void MainWindow::on_actionText_View_toggled(bool checked)
{
    if(checked) {
        stackedWidget_2->setCurrentIndex(1);
        currentDocument->refreshFileContent();
        textEdit->setCurrentFont(outputFont);
        textEdit->setText( currentDocument->fileContent() );
    }
    else
        stackedWidget_2->setCurrentIndex(0);

    updateUiTitle();
}

void MainWindow::on_actionBPP_triggered()
{
    stackedWidget->setCurrentIndex(BPP);
}

void MainWindow::on_actionOutput_Window_triggered()
{
    if(!bppOutput) return;

    bppOutput->show();

    if( bppOutput->isMinimized() )
        bppOutput->showNormal();
    if( !bppOutput->isActiveWindow() )
        bppOutput->activateWindow();
}

void MainWindow::on_actionConfiguration_triggered()
{
    ConfigurationDialog configurationDialog(this);
    if( configurationDialog.exec() == QDialog::Accepted ) {
        saveSettings();

        updateEnginesPath();
        updateOutputFont();
    }
}

void MainWindow::on_actionBPP_User_Guide_triggered()
{
    QString path = absoluteBPPPath + "/bppDOC.pdf";

    if( QFileInfo(path).isFile() ) {
#if defined(Q_OS_WIN32)
        QString program = "explorer.exe";
        QStringList arguments;
        arguments << QDir::toNativeSeparators(path);
        QProcess::startDetached(program, arguments);
#elif defined(Q_OS_MAC)
        QString program = "open";
        QStringList arguments;
        arguments << QDir::toNativeSeparators(path);
        QProcess::startDetached(program, arguments);
#elif defined(Q_OS_LINUX) && defined(Q_WS_X11)
        QString program = "okular";
        QStringList arguments;
        arguments << QDir::toNativeSeparators(path);
        QProcess::startDetached(program, arguments);
#else
#warning Nothing to do on current platform
#endif
    }
    else
        showWarningMessageWhenPathFailed(path, false);
}

void MainWindow::on_actionBPP_History_triggered()
{
    QString path = absoluteBPPPath + "/bppHistory.txt";

    if( QFileInfo(path).isFile() ) {
        BPPHistoryDialog bppHistoryDialog(path, outputFont, this);
        bppHistoryDialog.exec();
    }
    else
        showWarningMessageWhenPathFailed(path, false);
}

void MainWindow::on_actionHow_to_Cite_the_Program_triggered()
{
    QString citeString_1 = "Rannala, B., and Z. Yang. 2003. Bayes estimation of species divergence times and ancestral population sizes using DNA sequences from multiple loci. Genetics 164:1645-1656.";
    QString citeString_2 = "Yang, Z., and B. Rannala. 2010. Bayesian species delimitation using multilocus sequence data. Proc. Natl. Acad. Sci. U.S.A. 107:9264-9269.";

    QMessageBox msgBox;
    msgBox.setIcon(QMessageBox::NoIcon);
    msgBox.setWindowIcon(QIcon(":/bppxlogo.png"));
    msgBox.setWindowTitle(tr("How to Cite the Program"));
    msgBox.setText(tr("<p><b><font size=\"3\">You&nbsp;can&nbsp;cite&nbsp;the&nbsp;program&nbsp;as&nbsp;follows&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</font></b></p>"));
    msgBox.setInformativeText("<p><font face=\"Courier New\">" + citeString_1 + "</font></p>" + "<p><font face=\"Courier New\">" + citeString_2 + "</font></p>");
    QPushButton* copyToClipboardButton = msgBox.addButton(tr("Copy to Clipboard"), QMessageBox::AcceptRole);
    msgBox.addButton(QMessageBox::Cancel);
    msgBox.setDefaultButton(copyToClipboardButton);
    msgBox.exec();

    if( msgBox.clickedButton() == copyToClipboardButton ) {
        QApplication::clipboard()->setText(citeString_1 + "\n" + citeString_2);
    }
}

void MainWindow::on_actionAbout_BPP_X_triggered()
{
    QString appName = QApplication::applicationName();

    QMessageBox::about(this, tr("About %1").arg(appName),
                       tr("<p><b><font size=\"6\" face=\"Times\">%1</font></b>&nbsp;&nbsp;&nbsp;version %2</p>"
                          "<hr/>"
                          "<p>Based on Qt %3</p>"
                          "<p>%1&nbsp;is&nbsp;a&nbsp;graphical&nbsp;user&nbsp;interface&nbsp;for&nbsp;bpp&nbsp;program.</p>"
                          "<p>Copyright&nbsp;&copy;&nbsp;2012&nbsp;<b><font face=\"Times\">Bo&nbsp;Xu&nbsp;</font></b>&lt;<a href=\"mailto:xuxbob@gmail.com\">xuxbob@gmail.com</a>&gt;</p>").arg(appName).arg(QApplication::applicationVersion()).arg(QT_VERSION_STR)
                       );
}

void MainWindow::on_actionAbout_BPP_triggered()
{
    QMessageBox::about(this, tr("About BPP"),
                       tr("<p><b><font size=\"6\" face=\"Times\">Bayesian&nbsp;Phylogeny&nbsp;and&nbsp;Phylogeography</font></b></p>"
                          "<hr/>"
                          "<p>BPP is a program which implements Bayesian Markov chain Monte Carlo (MCMC) algorithms for analysis of phylogeographic data, that is, sequence alignments from multiple loci and multiple closely-related species.</p>"
                          "<p>It is available free for academic use only, written by <b><font face=\"Times\">Ziheng&nbsp;Yang&nbsp;</font></b>&lt;<a href=\"mailto:z.yang@ucl.ac.uk\">z.yang@ucl.ac.uk</a>&gt;.</p>")
                        );
}


//----- bpp -----

void MainWindow::on_pushButton_clicked()
{
    pushButton->setDisabled(true);

    if(!bppOutput) {
        bppOutput = new OutputWindow;
        bppOutput->setWindowTitle(tr("%1 Output").arg("BPP"));
        bppOutput->textEdit->setCurrentFont(outputFont);
        bppEngine->setOutputWindow(bppOutput->textEdit);
        QObject::connect(bppOutput, SIGNAL(closed()), bppEngine, SLOT(terminate()));
        QObject::connect(bppOutput, SIGNAL(deleted()), bppEngine, SLOT(removeOutputWindow()));
    }

    if( !bppEngine->run() ) {
        pushButton->setEnabled(true);
    }
    else {
        bppOutput->show();

        if( bppOutput->isMinimized() )
            bppOutput->showNormal();
        if( !bppOutput->isActiveWindow() )
            bppOutput->activateWindow();
    }
}

void MainWindow::on_pushButton_2_clicked()
{
    bppEngine->terminate();
}

void MainWindow::on_toolButton_clicked()
{
    QString path = QDir::cleanPath( comboBox_20->lineEdit()->text() );

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QDir::cleanPath( QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path) );

    QString fileName = QFileDialog::getOpenFileName(this, tr("Open seqfile"), path);
    if(fileName.isEmpty()) return;

    comboBox_20->lineEdit()->setText( QDir::toNativeSeparators(fileName) );
}

void MainWindow::on_toolButton_2_clicked()
{
    QString path = QDir::cleanPath( comboBox_21->lineEdit()->text() );

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QDir::cleanPath( QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path) );

    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Imapfile"), path);
    if(fileName.isEmpty()) return;

    comboBox_21->lineEdit()->setText( QDir::toNativeSeparators(fileName) );
}

void MainWindow::on_toolButton_3_clicked()
{
    QString path = QDir::cleanPath( comboBox_22->lineEdit()->text() );

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QDir::cleanPath( QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path) );

    QString directoryName = QFileDialog::getExistingDirectory(this, tr("Save output files in ..."), path);
    if(directoryName.isEmpty()) return;

    comboBox_22->lineEdit()->setText( QDir::toNativeSeparators(directoryName) );
}

void MainWindow::on_toolButton_4_clicked()
{
    QString path = QDir::cleanPath( comboBox_23->lineEdit()->text() );

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QDir::cleanPath( QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path) );

    QString fileName = QFileDialog::getOpenFileName(this, tr("Open locusratefile"), path);
    if(fileName.isEmpty()) return;

    comboBox_23->lineEdit()->setText( QDir::toNativeSeparators(fileName) );
}

void MainWindow::on_toolButton_5_clicked()
{
    QString path = QDir::cleanPath( comboBox_24->lineEdit()->text() );

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QDir::cleanPath( QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path) );

    QString fileName = QFileDialog::getOpenFileName(this, tr("Open heredityfile"), path);
    if(fileName.isEmpty()) return;

    comboBox_24->lineEdit()->setText( QDir::toNativeSeparators(fileName) );
}

void MainWindow::on_toolButton_6_clicked()
{
    QList<QTableWidgetSelectionRange> list = tableWidget->selectedRanges();
    QList<int> selectedRowList;
    bool hasOtherRange;
    getSelectedRows(list, selectedRowList, hasOtherRange);

    int row;
    if( !selectedRowList.isEmpty() && !hasOtherRange )
        row = selectedRowList.first();
    else
        row = tableWidget->rowCount();

    if( row > bppEngine->cList_species.size() ) return;
    static int count = 0;
    count++;
    if(count > 50) count -= 50;
    bppEngine->cList_species.insert(row, BPPEngine::Species(tr("%1%2").arg(tr("Unnamed-")).arg(count), 1));

    tableWidget->insertRow(row);
    tableWidget->setItem(row, 0, new QTableWidgetItem( bppEngine->cList_species[row].name ));
    tableWidget->setItem(row, 1, new QTableWidgetItem( QString::number(bppEngine->cList_species[row].nseqsp) ));
    lineEdit->setText( QString::number(tableWidget->rowCount()) );
}

void MainWindow::on_toolButton_7_clicked()
{
    QList<QTableWidgetSelectionRange> list = tableWidget->selectedRanges();
    QList<int> selectedRowList;
    bool hasOtherRange;
    getSelectedRows(list, selectedRowList, hasOtherRange);

    if( !selectedRowList.isEmpty() && !hasOtherRange ) {
        QList<int>::const_iterator it = selectedRowList.end() - 1;
        for(; it >= selectedRowList.begin(); --it) {
            tableWidget->removeRow(*it);
            bppEngine->cList_species.removeAt(*it);
        }
        lineEdit->setText( QString::number(tableWidget->rowCount()) );
    }
}

void MainWindow::on_tableWidget_itemSelectionChanged()
{
    QList<QTableWidgetSelectionRange> list = tableWidget->selectedRanges();
    QList<int> selectedRowList;
    bool hasOtherRange;
    getSelectedRows(list, selectedRowList, hasOtherRange);

    toolButton_7->setEnabled( !selectedRowList.isEmpty() && !hasOtherRange );
}

void MainWindow::on_checkBox_toggled(bool checked)
{
    label_18->setEnabled(!checked);
    label_19->setEnabled(!checked);
    label_20->setEnabled(!checked);
    label_29->setEnabled(!checked);
    label_97->setEnabled(!checked);
    label_98->setEnabled(!checked);
    label_108->setEnabled(!checked);
    doubleSpinBox_14->setEnabled(!checked);
    doubleSpinBox_15->setEnabled(!checked);
    doubleSpinBox_16->setEnabled(!checked);
    doubleSpinBox_17->setEnabled(!checked);
    doubleSpinBox_18->setEnabled(!checked);
    doubleSpinBox_19->setEnabled(!checked);
    doubleSpinBox_20->setEnabled(!checked);
}

void MainWindow::on_lineEdit_5_textEdited(const QString& text)
{
    bppEngine->nEdit_seed = text.toInt();
    setModifiedFlag();
}

void MainWindow::on_comboBox_activated(int index)
{
    int i = bppEngine->nComb_noisy;
    if(i > 4 && i < 9) i = 4;
    else if(i > 9) i = 9;

    if(index == 5) index = 9;

    if( i != index ) {
        bppEngine->nComb_noisy = index;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_20_editTextChanged(const QString& text)
{
    if( bppEngine->strEdit_seqfile != text ) {
        bppEngine->strEdit_seqfile = text;
        setModifiedFlag();
    }

    QString path = text;

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

    if( !QDir().exists(path) )
        comboBox_20->setStyleSheet("color: red");
    else
        comboBox_20->setStyleSheet("");
}

void MainWindow::on_comboBox_21_editTextChanged(const QString& text)
{
    if( bppEngine->strEdit_Imapfile != text ) {
        bppEngine->strEdit_Imapfile = text;
        setModifiedFlag();
    }

    if( comboBox_21->isEnabled() ) {
        QString path = text;

        if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
            path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

        if( !QDir().exists(path) )
            comboBox_21->setStyleSheet("color: red");
        else
            comboBox_21->setStyleSheet("");
    }
    else
        comboBox_21->setStyleSheet("");
}

void MainWindow::on_lineEdit_3_textEdited(const QString& text)
{
    bppEngine->strEdit_outfile_Name = text;
    setModifiedFlag();
}

void MainWindow::on_lineEdit_4_textEdited(const QString& text)
{
    bppEngine->strEdit_mcmcfile_Name = text;
    setModifiedFlag();
}

void MainWindow::on_comboBox_22_editTextChanged(const QString& text)
{
    if( bppEngine->strEdit_Output_Location != text ) {
        bppEngine->strEdit_Output_Location = text;
        setModifiedFlag();
    }

    QString path = text;

    if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
        path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

    if( !QDir(path).exists() )
        comboBox_22->setStyleSheet("color: red");
    else
        comboBox_22->setStyleSheet("");
}

void MainWindow::on_comboBox_2_currentIndexChanged(int index)
{
    label_26->setEnabled( index == 1 );
    comboBox_3->setEnabled( index == 1 );
    groupBox_10->setEnabled( index == 1 );

    int i = comboBox_8->currentIndex();
    label_34->setEnabled( index == 1 || i == 1 );
    comboBox_4->setEnabled( index == 1 || i == 1 );

    int n = comboBox_4->count();
    if(index == 1 && n == 2)
        comboBox_4->insertItems(2, QStringList()
         << QApplication::translate("MainWindow", "2: user probs", 0, QApplication::UnicodeUTF8)
        );
    else if(index == 0 && n == 3)
        comboBox_4->removeItem(2);

    if( bppEngine->nComb_speciesdelimitation != index ) {
        bppEngine->nComb_speciesdelimitation = index;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_3_currentIndexChanged(int index)
{
    label_5->setText(QApplication::translate("MainWindow", ((index == 1)? "\316\261":"\316\265"), 0, QApplication::UnicodeUTF8));
    label_6->setEnabled( index == 1 );
    doubleSpinBox_2->setEnabled( index == 1 );

    if( bppEngine->nComb_RJalgorithm != index ) {
        bppEngine->nComb_RJalgorithm = index;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_RJfinetune[0] - d) >= 0.000001 ) {
        bppEngine->nEdit_RJfinetune[0] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_2_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_RJfinetune[1] - d) >= 0.000001 ) {
        bppEngine->nEdit_RJfinetune[1] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_8_currentIndexChanged(int index)
{
    int i = comboBox_2->currentIndex();
    label_34->setEnabled( index == 1 || i == 1 );
    comboBox_4->setEnabled( index == 1 || i == 1 );

    if( bppEngine->nComb_speciestree != index ) {
        bppEngine->nComb_speciestree = index;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_4_activated(int index)
{
    if( bppEngine->nComb_speciesmodelprior != index ) {
        bppEngine->nComb_speciesmodelprior = index;
        setModifiedFlag();
    }
}

void MainWindow::on_lineEdit_textChanged(const QString& text)
{
    int i = text.toInt();

    bool b = comboBox_21->isEnabled();
    if(!b && i != 1) {
        QString path = comboBox_21->lineEdit()->text();

        if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
            path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

        if( !QDir().exists(path) )
            comboBox_21->setStyleSheet("color: red");
        else
            comboBox_21->setStyleSheet("");
    }
    else if(b && i == 1)
        comboBox_21->setStyleSheet("");

    label_2->setEnabled( i != 1 );
    comboBox_21->setEnabled( i != 1 );
    toolButton_2->setEnabled( i != 1 );

    if(i != 1) {
        if( !checkTreeFormat(lineEdit_6->text().trimmed(), i) )
            lineEdit_6->setStyleSheet("color: red");
        else
            lineEdit_6->setStyleSheet("");
    }
    else
        lineEdit_6->setStyleSheet("");

    label_28->setEnabled( i != 1 );
    lineEdit_6->setEnabled( i != 1 );

    if( bppEngine->nEdit_nspecies != i ) {
        bppEngine->nEdit_nspecies = i;
        setModifiedFlag();
    }
}

void MainWindow::on_tableWidget_itemChanged(QTableWidgetItem* item)
{
    int column = tableWidget->column(item);
    int row = tableWidget->row(item);

    if(column == 0) {
        QString name = item->text();
        if( row < bppEngine->nEdit_nspecies && bppEngine->cList_species[row].name != name ) {
            bppEngine->cList_species[row].name = name;
            setModifiedFlag();
        }
    }
    else if(column == 1) {
        int nseqsp = item->text().toInt();
        if( row < bppEngine->nEdit_nspecies && bppEngine->cList_species[row].nseqsp != nseqsp ) {
            bppEngine->cList_species[row].nseqsp = nseqsp;
            setModifiedFlag();
        }
    }
}

void MainWindow::on_lineEdit_6_textChanged(const QString& text)
{
    if( bppEngine->strEdit_tree != text ) {
        bppEngine->strEdit_tree = text;
        setModifiedFlag();
    }

    if( lineEdit_6->isEnabled() ) {
        if( !checkTreeFormat(text.trimmed(), lineEdit->text().toInt()) )
            lineEdit_6->setStyleSheet("color: red");
        else
            lineEdit_6->setStyleSheet("");
    }
    else
        lineEdit_6->setStyleSheet("");
}

void MainWindow::on_comboBox_5_activated(int index)
{
    if( bppEngine->nComb_usedata != index ) {
        bppEngine->nComb_usedata = index;
        setModifiedFlag();
    }
}

void MainWindow::on_spinBox_2_valueChanged(int i)
{
    if( bppEngine->nEdit_nloci != i ) {
        bppEngine->nEdit_nloci = i;
        setModifiedFlag();
    }
}

void MainWindow::on_checkBox_2_stateChanged(int state)
{
    if( bppEngine->bButt_cleandata != bool(state) ) {
        bppEngine->bButt_cleandata = state;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_6_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_thetaprior[0] - d) >= 0.000001 ) {
        bppEngine->nEdit_thetaprior[0] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_7_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_thetaprior[1] - d) >= 0.000001 ) {
        bppEngine->nEdit_thetaprior[1] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_8_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_tauprior[0] - d) >= 0.000001 ) {
        bppEngine->nEdit_tauprior[0] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_9_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_tauprior[1] - d) >= 0.000001 ) {
        bppEngine->nEdit_tauprior[1] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_10_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_tauprior[2] - d) >= 0.000001 ) {
        bppEngine->nEdit_tauprior[2] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_6_currentIndexChanged(int index)
{
    groupBox_14->setEnabled( index == 1 );

    bool b = comboBox_23->isEnabled();
    if(!b && index == 2) {
        QString path = comboBox_23->lineEdit()->text();

        if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
            path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

        if( !QDir().exists(path) )
            comboBox_23->setStyleSheet("color: red");
        else
            comboBox_23->setStyleSheet("");
    }
    else if(b && index != 2)
        comboBox_23->setStyleSheet("");

    label_39->setEnabled( index == 2 );
    comboBox_23->setEnabled( index == 2 );
    toolButton_4->setEnabled( index == 2 );

    if( bppEngine->nComb_locusrate != index ) {
        bppEngine->nComb_locusrate = index;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_11_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_a_locusrate - d) >= 0.000001 ) {
        bppEngine->nEdit_a_locusrate = d;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_23_editTextChanged(const QString& text)
{
    if( bppEngine->strEdit_locusratefile != text ) {
        bppEngine->strEdit_locusratefile = text;
        setModifiedFlag();
    }

    if( comboBox_23->isEnabled() ) {
        QString path = text;

        if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
            path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

        if( !QDir().exists(path) )
            comboBox_23->setStyleSheet("color: red");
        else
            comboBox_23->setStyleSheet("");
    }
    else
        comboBox_23->setStyleSheet("");
}

void MainWindow::on_comboBox_7_currentIndexChanged(int index)
{
    groupBox_15->setEnabled( index == 1 );

    bool b = comboBox_24->isEnabled();
    if(!b && index == 2) {
        QString path = comboBox_24->lineEdit()->text();

        if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
            path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

        if( !QDir().exists(path) )
            comboBox_24->setStyleSheet("color: red");
        else
            comboBox_24->setStyleSheet("");
    }
    else if(b && index != 2)
        comboBox_24->setStyleSheet("");

    label_40->setEnabled( index == 2 );
    comboBox_24->setEnabled( index == 2 );
    toolButton_5->setEnabled( index == 2 );

    if( bppEngine->nComb_heredity != index ) {
        bppEngine->nComb_heredity = index;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_12_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_a_heredity - d) >= 0.000001 ) {
        bppEngine->nEdit_a_heredity = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_13_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_b_heredity - d) >= 0.000001 ) {
        bppEngine->nEdit_b_heredity = d;
        setModifiedFlag();
    }
}

void MainWindow::on_comboBox_24_editTextChanged(const QString& text)
{
    if( bppEngine->strEdit_heredityfile != text ) {
        bppEngine->strEdit_heredityfile = text;
        setModifiedFlag();
    }

    if( comboBox_24->isEnabled() ) {
        QString path = text;

        if( !QDir::isAbsolutePath(path) && currentDocument && !currentDocument->fileName().isEmpty() )
            path = QFileInfo( currentDocument->fileName() ).absoluteDir().absoluteFilePath(path);

        if( !QDir().exists(path) )
            comboBox_24->setStyleSheet("color: red");
        else
            comboBox_24->setStyleSheet("");
    }
    else
        comboBox_24->setStyleSheet("");
}

void MainWindow::on_checkBox_stateChanged(int state)
{
    if( bppEngine->bButt_autoAdjustFinetune != bool(state) ) {
        bppEngine->bButt_autoAdjustFinetune = state;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_14_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[0] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[0] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_15_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[1] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[1] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_16_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[2] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[2] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_17_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[3] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[3] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_18_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[4] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[4] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_19_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[5] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[5] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_doubleSpinBox_20_valueChanged(double d)
{
    if( ::qAbs(bppEngine->nEdit_finetune[6] - d) >= 0.000001 ) {
        bppEngine->nEdit_finetune[6] = d;
        setModifiedFlag();
    }
}

void MainWindow::on_checkBox_3_stateChanged(int state)
{
    if( bppEngine->bButt_print != bool(state) ) {
        bppEngine->bButt_print = state;
        setModifiedFlag();
    }
}

void MainWindow::on_checkBox_4_stateChanged(int state)
{
    if( bppEngine->bButt_printlocusrate != bool(state) ) {
        bppEngine->bButt_printlocusrate = state;
        setModifiedFlag();
    }
}

void MainWindow::on_checkBox_5_stateChanged(int state)
{
    if( bppEngine->bButt_printheredity != bool(state) ) {
        bppEngine->bButt_printheredity = state;
        setModifiedFlag();
    }
}

void MainWindow::on_checkBox_6_stateChanged(int state)
{
    if( bppEngine->bButt_printGenetree != bool(state) ) {
        bppEngine->bButt_printGenetree = state;
        setModifiedFlag();
    }
}

void MainWindow::on_spinBox_3_valueChanged(int i)
{
    if( bppEngine->nEdit_burnin != i ) {
        bppEngine->nEdit_burnin = i;
        setModifiedFlag();
    }
}

void MainWindow::on_spinBox_4_valueChanged(int i)
{
    if( bppEngine->nEdit_sampfreq != i ) {
        bppEngine->nEdit_sampfreq = i;
        setModifiedFlag();
    }
}

void MainWindow::on_spinBox_5_valueChanged(int i)
{
    if( bppEngine->nEdit_nsample != i ) {
        bppEngine->nEdit_nsample = i;
        setModifiedFlag();
    }
}
