//------------------------------------------------
// Copyright (C) 2012 Bo Xu <xuxbob@gmail.com>
//------------------------------------------------

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "ui_mainwindow.h"

#include "engine.h"

class BPPEngine;
class OutputWindow;

class MainWindow : public QMainWindow, private Ui::MainWindow
{
    Q_OBJECT

public:
    enum {
        StartPage = 0,
        BPP
    };

public:
    explicit MainWindow(QWidget *parent = 0);
    virtual ~MainWindow();

protected:
    bool eventFilter(QObject* watched, QEvent* event);
    void closeEvent(QCloseEvent* event);
    void dragEnterEvent(QDragEnterEvent* event);
    void dropEvent(QDropEvent* event);

private:
    void loadSettings();
    void saveSettings();
    void loadDocumentFile(const QString& fileName);
    bool saveDocumentFile(const QString& fileName);
    void addToRecentFiles(const QString& fileName);
    void removeFromRecentFiles(const QString& fileName);
    bool saveModifiedFile(Document* document, bool saveInBackground = false);
    void setModifiedFlag();
    void updateEnginesPath();
    void updateOutputFont();
    void updateRecentFileActions();
    void updateUiTitle();
    void updateUi();
    void updateEngine(); //for dubug only
    void showWarningMessageWhenPathFailed(const QString& aPath, bool isExecutable);

    friend class ConfigurationDialog;

private:
    void getSelectedRows(QList<QTableWidgetSelectionRange>& list, QList<int>& selectedRowList, bool& hasOtherRange) const;
    bool checkTreeFormat(const QString& tree, int n) const;

public slots:
    void onBPPEngineStateChanged(Engine::EngineState newState);

private slots:
    void print(QPrinter* printer);
    void on_stackedWidget_currentChanged(int index);

    //menu
    void on_menuFile_aboutToShow();
    void on_actionNew_triggered();
    void on_actionOpen_triggered();
    void on_actionClose_triggered();
    bool on_actionSave_triggered();
    bool on_actionSave_As_triggered();
    void on_actionSet_Printer_Font_triggered();
    void on_actionPrint_Preview_triggered();
    void on_actionPrint_triggered();
    void onActionOpenRecentFile();
    void on_actionClear_History_triggered();

    void on_menuEdit_aboutToShow();
    void on_actionUndo_triggered();
    void on_actionRedo_triggered();
    void on_actionCut_triggered();
    void on_actionCopy_triggered();
    void on_actionPaste_triggered();
    void on_actionDelete_triggered();
    void on_actionSelect_All_triggered();

    void on_menuFunction_aboutToShow();
    void on_actionStart_Page_triggered();
    void on_actionBPP_triggered();
    void on_actionText_View_toggled(bool checked);
    void on_actionOutput_Window_triggered();
    void on_actionConfiguration_triggered();

    void on_actionBPP_User_Guide_triggered();
    void on_actionBPP_History_triggered();
    void on_actionHow_to_Cite_the_Program_triggered();
    void on_actionAbout_BPP_X_triggered();
    void on_actionAbout_BPP_triggered();

    //bpp
    void on_pushButton_clicked();
    void on_pushButton_2_clicked();
    void on_toolButton_clicked();
    void on_toolButton_2_clicked();
    void on_toolButton_3_clicked();
    void on_toolButton_4_clicked();
    void on_toolButton_5_clicked();
    void on_toolButton_6_clicked();
    void on_toolButton_7_clicked();
    void on_tableWidget_itemSelectionChanged();
    void on_checkBox_toggled(bool checked);

    void on_lineEdit_5_textEdited(const QString& text);
    void on_comboBox_activated(int index);
    void on_comboBox_20_editTextChanged(const QString& text);
    void on_comboBox_21_editTextChanged(const QString& text);
    void on_lineEdit_3_textEdited(const QString& text);
    void on_lineEdit_4_textEdited(const QString& text);
    void on_comboBox_22_editTextChanged(const QString& text);
    void on_comboBox_2_currentIndexChanged(int index);
    void on_comboBox_3_currentIndexChanged(int index);
    void on_doubleSpinBox_valueChanged(double d);
    void on_doubleSpinBox_2_valueChanged(double d);
    void on_comboBox_8_currentIndexChanged(int index);
    void on_comboBox_4_activated(int index);
    void on_lineEdit_textChanged(const QString& text);
    void on_tableWidget_itemChanged(QTableWidgetItem* item);
    void on_lineEdit_6_textChanged(const QString& text);
    void on_comboBox_5_activated(int index);
    void on_spinBox_2_valueChanged(int i);
    void on_checkBox_2_stateChanged(int state);
    void on_doubleSpinBox_6_valueChanged(double d);
    void on_doubleSpinBox_7_valueChanged(double d);
    void on_doubleSpinBox_8_valueChanged(double d);
    void on_doubleSpinBox_9_valueChanged(double d);
    void on_doubleSpinBox_10_valueChanged(double d);
    void on_comboBox_6_currentIndexChanged(int index);
    void on_doubleSpinBox_11_valueChanged(double d);
    void on_comboBox_23_editTextChanged(const QString& text);
    void on_comboBox_7_currentIndexChanged(int index);
    void on_doubleSpinBox_12_valueChanged(double d);
    void on_doubleSpinBox_13_valueChanged(double d);
    void on_comboBox_24_editTextChanged(const QString& text);
    void on_checkBox_stateChanged(int state);
    void on_doubleSpinBox_14_valueChanged(double d);
    void on_doubleSpinBox_15_valueChanged(double d);
    void on_doubleSpinBox_16_valueChanged(double d);
    void on_doubleSpinBox_17_valueChanged(double d);
    void on_doubleSpinBox_18_valueChanged(double d);
    void on_doubleSpinBox_19_valueChanged(double d);
    void on_doubleSpinBox_20_valueChanged(double d);
    void on_checkBox_3_stateChanged(int state);
    void on_checkBox_4_stateChanged(int state);
    void on_checkBox_5_stateChanged(int state);
    void on_checkBox_6_stateChanged(int state);
    void on_spinBox_3_valueChanged(int i);
    void on_spinBox_4_valueChanged(int i);
    void on_spinBox_5_valueChanged(int i);

private:
    QString                   bppPath;
    QFont                     outputFont;
    QFont                     printFont;
    QStringList               recentFiles;
    QString                   lastFileOpenPath;
    QString                   absoluteBPPPath;
    const QFont               defaultOutputFont;
    const QFont               defaultPrintFont;

    BPPEngine*                bppEngine;
    OutputWindow*             bppOutput;

    Document*                 currentDocument;

    enum {
        MaxRecentFiles = 10
    };

    QAction*                  recentFileActions[MaxRecentFiles];
    QWidget*                  focusEdit;
    QActionGroup*             actionGroup;
};

#endif // MAINWINDOW_H
